try:
    import streamlit as st
except ImportError:
    print("Please install streamlit - python3 -m pip install streamlit")
    exit(1)

try:
    import fenics_plotly
except ImportError:
    print("Please install fenics_plotly - python3 -m pip install fenics_plotly")
    exit(1)


from pathlib import Path
import matplotlib.pyplot as plt


import simcardems
import mpi4py

import pulse
import hashlib

simcardems_folder = Path.home().joinpath("simcardems")


def return_none(*args, **kwargs):
    return None


def about():

    st.title("About")

    for i in range(8):
        st.sidebar.write("")

    intro_markdown = """
    The SIMula CARDiac ElectroMechanics solver
    is a FEniCS-based cardiac electro-mechanics
    solver and is developed as a part of the
    [SimCardio Test project](https://www.simcardiotest.eu/wordpress).
    The solver depends on the mechanics solver
    [pulse](https://github.com/ComputationalPhysiology/pulse)
    and electrophysiology solver [cbcbeat](https://github.com/ComputationalPhysiology/cbcbeat).

    Please consult the [documentation](http://computationalphysiology.github.io/simcardems)
    if you want to learn more.
    """
    st.markdown(intro_markdown)

    return


class Postprocess:
    @staticmethod
    def load_data(outdir, reduction="center"):

        loader = simcardems.DataLoader(outdir.joinpath("results.h5"))
        return simcardems.postprocess.extract_traces(loader=loader, reduction=reduction)

    @staticmethod
    def make_figures(values):
        fig, ax = plt.subplots(2, 2, figsize=(10, 8), sharex=True)
        ax[0, 0].plot(values["time"][1:], values["mechanics"]["lmbda"][1:])
        ax[0, 1].plot(values["time"][1:], values["mechanics"]["Ta"][1:])
        ax[1, 0].plot(values["time"], values["ep"]["V"])
        ax[1, 1].plot(values["time"][1:], values["ep"]["Ca"][1:])

        ax[0, 0].set_title(r"$\lambda$")
        ax[0, 1].set_title("Ta")
        ax[1, 0].set_title("V")
        ax[1, 1].set_title("Ca")
        for axi in ax.flatten():
            axi.grid()
            if False:
                axi.set_xlim([0, 5000])
        ax[1, 0].set_xlabel("Time [ms]")
        ax[1, 1].set_xlabel("Time [ms]")
        ax[0, 0].set_ylim(
            [
                min(0.9, min(values["mechanics"]["lmbda"][1:])),
                max(1.1, max(values["mechanics"]["lmbda"][1:])),
            ],
        )
        return fig


def postprocess():
    st.title("Postprocessing")

    for i in range(8):
        st.sidebar.write("")

    intro_markdown = f"""
    Here is the postprocessing page.

    Here you will be able to
    load all the results that are located in the folder
    `{simcardems_folder}`

    """
    st.markdown(intro_markdown)
    if not simcardems_folder.exists():
        st.info("No results found")
        return

    outdir = st.selectbox(
        "Select result directory",
        [""] + list(simcardems_folder.iterdir()),
        format_func=lambda x: "Select an option" if x == "" else x,
    )
    if outdir == "":
        st.info("Please select a result directory")
        return

    outdir = Path(outdir)

    # TODO: Use multiselect instead
    reduction = st.selectbox(
        "Select a reduction",
        [""] + ["center", "average"],
        format_func=lambda x: "Select an option" if x == "" else x,
    )
    if reduction == "":
        st.info("Please select a node")
        return

    values = Postprocess.load_data(outdir, reduction=reduction)
    fig = Postprocess.make_figures(values)
    st.pyplot(fig)

    return


class Simulation:
    @staticmethod
    def plot_mesh(mesh, label):
        fig_mech_mesh = fenics_plotly.fenics_plotly.plot(mesh, show=False)
        st.subheader(f"{label} mesh")
        st.text(f"Num cells: {mesh.num_cells()}, Num vertices: {mesh.num_vertices()}")
        st.plotly_chart(fig_mech_mesh.figure)

    @staticmethod
    def handle_mesh():
        st.header("Mesh")
        cols_l_xyz = st.columns(3)
        params = simcardems.geometry.SlabGeometry.default_parameters()
        with cols_l_xyz[0]:
            lx = st.number_input("lx", value=params["lx"])
        with cols_l_xyz[1]:
            ly = st.number_input("ly", value=params["ly"])
        with cols_l_xyz[2]:
            lz = st.number_input("lz", value=params["lz"])

        cols_dx_ref = st.columns(2)
        with cols_dx_ref[0]:
            dx = st.number_input("dx", value=params["dx"])
        with cols_dx_ref[1]:
            num_refinements = st.number_input(
                "num_refinements",
                value=params["num_refinements"],
            )

        return {
            "lx": lx,
            "ly": ly,
            "lz": lz,
            "dx": dx,
            "num_refinements": num_refinements,
        }

    @staticmethod
    def handle_ep_solver_args():
        st.header("EP solver")

        cols_ep = st.columns(2)
        with cols_ep[0]:
            dt = st.number_input(
                "dt",
                value=simcardems.Config.dt,
                help="Time step for EP solver",
            )
        with cols_ep[1]:
            disease_state = st.selectbox(
                "Disease state",
                ["healthy", "hf"],
                help="Healthy or heart failure",
            )

        cols_files = st.columns(3)
        with cols_files[0]:
            cell_init_file = st.file_uploader(
                "Cell initial conditions",
                type=["json", "h5"],
                help="File with initial conditions for the cell",
            )
        with cols_files[1]:
            drug_factors_file = st.file_uploader(
                "Drug factors",
                type="json",
                help="File with drug factors",
            )
        with cols_files[2]:
            popu_factors_file = st.file_uploader(
                "Population factors",
                type="json",
                help="File with population factors",
            )
        return {
            "dt": dt,
            "disease_state": disease_state,
            "cell_init_file": cell_init_file,
            "drug_factors_file": drug_factors_file,
            "popu_factors_file": popu_factors_file,
        }

    @staticmethod
    def handle_mechanics_args():
        st.header("Mechanics solver")
        cols_mech = st.columns(4)
        with cols_mech[0]:
            bnd_rigid = st.checkbox("Rigid motion conditions")
        with cols_mech[1]:
            add_pre_stretch = st.checkbox("Add pre stretch")
            pre_stretch = None
            if add_pre_stretch:
                pre_stretch = st.number_input("Pre stretch value", 0)

        with cols_mech[2]:
            add_traction = st.checkbox("Add pre traction")
            traction = None
            if add_traction:
                traction = st.number_input("Traction value", 0)

        with cols_mech[3]:
            add_spring = st.checkbox("Add spring")
            spring = None
            if add_spring:
                spring = st.number_input("Spring value", 0)

        return {
            "bnd_rigid": bnd_rigid,
            "pre_stretch": pre_stretch,
            "traction": traction,
            "spring": spring,
        }

    @staticmethod
    @st.cache(
        suppress_st_warning=True,
        allow_output_mutation=True,
        show_spinner=True,
        hash_funcs={mpi4py.MPI.Op: return_none},
    )
    def load_model(geometry_args, ep_solver_args, mechanics_args):

        st.info("Create geometry")
        geometry = simcardems.geometry.SlabGeometry(parameters=geometry_args)
        st.info("Create EM coupling")
        coupling = simcardems.EMCoupling(geometry)
        st.info("Create EP model")
        ep_solver = simcardems.setup_models.setup_ep_solver(
            coupling=coupling, **ep_solver_args
        )
        coupling.register_ep_model(ep_solver)
        st.info("Create Mechanics model")
        mech_heart: pulse.MechanicsProblem = (
            simcardems.setup_models.setup_mechanics_solver(
                coupling=coupling,
                cell_params=ep_solver.ode_solver._model.parameters(),
                linear_solver="superlu_dist",
                geo=geometry,
                **mechanics_args,
            )
        )
        st.success("Done loading model")
        return simcardems.Runner.from_models(
            coupling=coupling,
            ep_solver=ep_solver,
            mech_heart=mech_heart,
            geo=geometry,
        )

    @staticmethod
    def visualize_model(
        coupling,
    ):
        st.header("Visualize model")
        cols_plot_mesh = st.columns(2)
        with cols_plot_mesh[0]:
            plot_mechanics_mesh = st.checkbox("Plot mechanics mesh")
        with cols_plot_mesh[1]:
            plot_ep_mesh = st.checkbox("Plot EP mesh")

        if plot_mechanics_mesh:
            Simulation.plot_mesh(coupling.geometry.mechanics_mesh, "Mechanics")

        if plot_ep_mesh:
            Simulation.plot_mesh(coupling.geometry.ep_mesh, "EP")

    @staticmethod
    def setup_simulation(geometry_args, ep_solver_args, mechanics_args):
        st.header("Run")

        cols_run = st.columns(3)
        with cols_run[0]:
            T = st.number_input(
                "T",
                value=simcardems.Config.T,
                help="End time",
            )

        with cols_run[1]:
            save_freq = st.number_input(
                "Save frequency",
                value=simcardems.Config.save_freq,
                help="How often to save the results",
            )

        all_options = {"T": T, **geometry_args, **ep_solver_args, **mechanics_args}
        simulation_id = hashlib.md5(repr(all_options).encode()).hexdigest()
        with cols_run[2]:
            outdir = st.text_input(
                "Output directory",
                value=simcardems_folder.joinpath(f"results_{simulation_id}"),
            )
        outdir = Path(outdir)
        create_result_dir = st.checkbox(f"Create result directory '{outdir}'")
        if not create_result_dir:
            return T, save_freq, None

        if not outdir.is_dir():
            st.info(f"Directory '{outdir}' does not exist. Creating...")
            outdir.mkdir(parents=True)
        else:
            st.info(f"Directory '{outdir}' already exist")
        return T, save_freq, outdir


def simulation():
    st.title("Simulation")
    geometry_args = Simulation.handle_mesh()
    ep_solver_args = Simulation.handle_ep_solver_args()
    mechanics_args = Simulation.handle_mechanics_args()

    load_model = st.checkbox("Load model")
    if not load_model:
        return

    runner: simcardems.Runner = Simulation.load_model(
        geometry_args=geometry_args,
        ep_solver_args=ep_solver_args,
        mechanics_args=mechanics_args,
    )
    Simulation.visualize_model(coupling=runner.coupling)

    T, save_freq, outdir = Simulation.setup_simulation(
        geometry_args,
        ep_solver_args,
        mechanics_args,
    )
    if outdir is None:
        return

    if st.button("Run simulation"):
        runner.outdir = outdir
        st_progress = st.progress(0)
        runner.solve(T=T, save_freq=save_freq, st_progress=st_progress)
        st.success("Done!")

    return


# Page settings
st.set_page_config(page_title="simcardems")

# Sidebar settings
pages = {
    "About": about,
    "Simulation": simulation,
    "Postprocess": postprocess,
}

st.sidebar.title("Simcardems")

# Radio buttons to select desired option
page = st.sidebar.radio("Pages", tuple(pages.keys()))

pages[page]()

# About
st.sidebar.markdown(
    """
- [Source code](https://github.com/ComputationalPhysiology/simcardems)
- [Documentation](http://computationalphysiology.github.io/simcardems)
""",
)



import json
from pathlib import Path
from typing import Dict
from typing import Optional

import cbcbeat
import dolfin
import pulse
import ufl

from . import utils
from .ORdmm_Land import ORdmm_Land as CellModel

logger = utils.getLogger(__name__)


def default_conductivities():
    # Conductivities as defined by page 4339 of Niederer benchmark
    return dict(
        sigma_il=0.17,  # mS / mm
        sigma_it=0.019,  # mS / mm
        sigma_el=0.62,  # mS / mm
        sigma_et=0.24,  # mS / mm
    )


def define_conductivity_tensor(
    chi: float = 140.0,
    C_m: float = 0.01,
    microstructure: Optional[pulse.Microstructure] = None,
):

    if microstructure is None:
        microstructure = default_microstructure()

    fiber = microstructure.f0
    sheet = microstructure.s0
    cross_sheet = microstructure.n0

    A = dolfin.as_matrix(
        [
            [fiber[0], sheet[0], cross_sheet[0]],
            [fiber[1], sheet[1], cross_sheet[1]],
            [fiber[2], sheet[2], cross_sheet[2]],
        ],
    )

    # FIXME: Make is possible to change this from the outside
    conductivities = default_conductivities()

    # Compute monodomain approximation by taking harmonic mean in each
    # direction of intracellular and extracellular part
    def harmonic_mean(a, b):
        return a * b / (a + b)

    sigma_l = harmonic_mean(conductivities["sigma_il"], conductivities["sigma_el"])
    sigma_t = harmonic_mean(conductivities["sigma_it"], conductivities["sigma_et"])

    # Scale conductivities by 1/(C_m * chi)
    s_l = sigma_l / (C_m * chi)  # mm^2 / ms
    s_t = sigma_t / (C_m * chi)  # mm^2 / ms

    # Define conductivity tensor
    M_star = ufl.diag(dolfin.as_vector([s_l, s_t, s_t]))
    M = A * M_star * A.T

    return M


def default_microstructure():
    return pulse.Microstructure(
        f0=dolfin.as_vector([1.0, 0.0, 0.0]),
        s0=dolfin.as_vector([0.0, 1.0, 0.0]),
        n0=dolfin.as_vector([0.0, 0.0, 1.0]),
    )


def setup_ep_model(
    cellmodel,
    mesh,
    PCL=1000,
    microstructure: Optional[pulse.Microstructure] = None,
):
    """Set-up cardiac model based on benchmark parameters."""

    if microstructure is None:
        microstructure = default_microstructure()

    # Define time
    time = dolfin.Constant(0.0)

    # Surface to volume ratio
    chi = 140.0  # mm^{-1}
    # Membrane capacitance
    C_m = 0.01  # mu F / mm^2

    # Define conductivity tensor
    M = define_conductivity_tensor(chi, C_m, microstructure=microstructure)

    # Mark stimulation region defined as [0, L]^3
    S1_marker = 1
    S1_markers = dolfin.MeshFunction("size_t", mesh, mesh.topology().dim())
    S1_markers.set_all(S1_marker)  # Mark the whole mesh

    # Define stimulation (NB: region of interest carried by the mesh
    # and assumptions in cbcbeat)
    duration = 2.0  # ms
    A = 50000.0  # mu A/cm^3
    cm2mm = 10.0
    factor = 1.0 / (chi * C_m)  # NB: cbcbeat convention
    amplitude = factor * A * (1.0 / cm2mm) ** 3  # mV/ms

    s = "((std::fmod(time,PCL) >= start) & (std::fmod(time,PCL) <= duration + start)) ? amplitude : 0.0"

    I_s = dolfin.Expression(
        s,
        time=time,
        start=0.0,
        duration=duration,
        amplitude=amplitude,
        PCL=PCL,
        degree=0,
    )
    # Store input parameters in cardiac model
    stimulus = cbcbeat.Markerwise((I_s,), (S1_marker,), S1_markers)

    petsc_options = [
        ["ksp_type", "cg"],
        ["pc_type", "gamg"],
        ["pc_gamg_verbose", "10"],
        ["pc_gamg_square_graph", "0"],
        ["pc_gamg_coarse_eq_limit", "3000"],
        ["mg_coarse_pc_type", "redundant"],
        ["mg_coarse_sub_pc_type", "lu"],
        ["mg_levels_ksp_type", "richardson"],
        ["mg_levels_ksp_max_it", "3"],
        ["mg_levels_pc_type", "sor"],
    ]
    for opt in petsc_options:
        dolfin.PETScOptions.set(*opt)

    heart = cbcbeat.CardiacModel(
        domain=mesh,
        time=time,
        M_i=M,
        M_e=None,
        cell_models=cellmodel,
        stimulus=stimulus,
        applied_current=None,
    )

    return heart


def setup_splitting_solver_parameters(
    dt,
    theta=0.5,
    preconditioner="sor",
    scheme="GRL1",
):
    ps = cbcbeat.SplittingSolver.default_parameters()
    ps["pde_solver"] = "monodomain"
    ps["MonodomainSolver"]["linear_solver_type"] = "iterative"
    ps["MonodomainSolver"]["theta"] = theta
    ps["MonodomainSolver"]["preconditioner"] = preconditioner
    ps["MonodomainSolver"]["default_timestep"] = dt
    ps["MonodomainSolver"]["use_custom_preconditioner"] = False
    ps["theta"] = theta
    ps["enable_adjoint"] = False
    ps["apply_stimulus_current_to_pde"] = True
    # ps["BasicCardiacODESolver"]["scheme"] = scheme
    ps["CardiacODESolver"]["scheme"] = scheme
    # ps["ode_solver_choice"] = "BasicCardiacODESolver"
    # ps["BasicCardiacODESolver"]["V_polynomial_family"] = "CG"
    # ps["BasicCardiacODESolver"]["V_polynomial_degree"] = 1
    # ps["BasicCardiacODESolver"]["S_polynomial_family"] = "CG"
    # ps["BasicCardiacODESolver"]["S_polynomial_degree"] = 1
    return ps


def file_exist(filename: Optional[str], suffix: str) -> bool:
    return (
        filename is not None
        and filename != ""
        and Path(filename).is_file()
        and Path(filename).suffix == suffix
    )


def load_json(filename: str):
    with open(filename, "r") as fid:
        d = json.load(fid)
    return d


def handle_cell_params(
    cell_params: Optional[Dict[str, float]] = None,
    disease_state: str = "healthy",
    drug_factors_file: str = "",
    popu_factors_file: str = "",
):
    cell_params_tmp = CellModel.default_parameters(disease_state)
    # FIXME: In this case we update the parameters first, while in the
    # initial condition case we do that last. We need to be consistent
    # about this.
    if cell_params is not None:
        cell_params_tmp.update(cell_params)
    # Adding optional drug factors to parameters (if drug_factors_file exists)
    if file_exist(drug_factors_file, ".json"):
        logger.info(f"Drug scaling factors loaded from {drug_factors_file}")
        cell_params_tmp.update(load_json(drug_factors_file))
    else:
        if drug_factors_file != "":
            logger.warning(f"Unable to load drug factors file {drug_factors_file}")
    # FIXME: A problem here is that popu_factors_file will overwrite the
    # drug_factors_file. Is it possible to only have one file?
    if file_exist(popu_factors_file, ".json"):
        logger.info(f"Population scaling factors loaded from {popu_factors_file}")
        cell_params_tmp.update(load_json(popu_factors_file))
    else:
        if popu_factors_file != "":
            logger.warning(f"Unable to load popu factors file {popu_factors_file}")

    return cell_params_tmp


def handle_cell_inits(
    cell_inits: Optional[Dict[str, float]] = None,
    cell_init_file: str = "",
) -> Dict[str, float]:
    cell_inits_tmp = CellModel.default_initial_conditions()
    if file_exist(cell_init_file, ".json"):
        cell_inits_tmp.update(load_json(cell_init_file))

    if file_exist(cell_init_file, ".h5"):
        cell_inits_tmp.update(load_json(cell_init_file))
        from .save_load_functions import load_initial_conditions_from_h5

        cell_inits = load_initial_conditions_from_h5(cell_init_file)

    # FIXME: This is a bit confusing, since it will overwrite the
    # inputs from the cell_init_file. There should be only one way to
    # do this IMO. I think this might be difficult for the user to reason
    # about. I think in general we should handle the loading from files
    # at higher level.
    if cell_inits is not None:
        cell_inits_tmp.update(cell_inits)
    return cell_inits_tmp



"""This module contains a ORdmm_Land_em_coupling cardiac cell model

The module was autogenerated from a gotran ode file
"""
from __future__ import division

from collections import OrderedDict
from typing import Dict

import dolfin
import ufl
from cbcbeat.cellmodels import CardiacCellModel
from dolfin import as_vector
from dolfin import Constant

from . import utils

logger = utils.getLogger(__name__)


def Max(a, b):
    return (a + b + abs(a - b)) / Constant(2.0)


def Min(a, b):
    return (a + b - abs(a - b)) / Constant(2.0)


def vs_functions_to_dict(vs):
    state_names = ORdmm_Land.default_initial_conditions().keys()
    return {
        name: utils.sub_function(vs, index) for index, name in enumerate(state_names)
    }


class ORdmm_Land(CardiacCellModel):
    def __init__(self, params=None, init_conditions=None):
        """
        Create cardiac cell model

        *Arguments*
         params (dict, :py:class:`dolfin.Mesh`, optional)
           optional model parameters
         init_conditions (dict, :py:class:`dolfin.Mesh`, optional)
           optional initial conditions
        """

        super().__init__(params, init_conditions)

    @staticmethod
    def default_parameters(disease_state: str = "healthy") -> Dict[str, float]:
        """Set-up and return default parameters.

        Parameters
        ----------
        disease_state : str, optional
            String with "hf" or "healthy", by default "healthy".
            If "hf", then parameters representing heart failure
            will be used.

        Returns
        -------
        OrderedDict
            Dictionary with default values
        """
        params = OrderedDict(
            [
                ("scale_ICaL", 1.018),
                ("scale_IK1", 1.414),
                ("scale_IKr", 1.119),
                ("scale_IKs", 1.648),
                ("scale_INaL", 2.274),
                ("celltype", 0),
                ("cao", 1.8),
                ("ko", 5.4),
                ("nao", 140.0),
                ("F", 96485.0),
                ("R", 8314.0),
                ("T", 310.0),
                ("L", 0.01),
                ("rad", 0.0011),
                ("Ahf", 0.99),
                ("GNa", 31),
                ("thL", 200.0),
                ("Gto", 0.02),
                ("delta_epi", 1.0),
                ("Aff", 0.6),
                ("Kmn", 0.002),
                ("k2n", 1000.0),
                ("tjca", 75.0),
                ("zca", 2.0),
                ("bt", 4.75),
                ("Beta0", 2.3),
                ("Beta1", -2.4),
                ("Tot_A", 25),
                ("Tref", 120),
                ("Trpn50", 0.35),
                ("calib", 1),
                ("cat50_ref", 0.805),
                ("emcoupling", 1),
                ("etal", 200),
                ("etas", 20),
                ("gammas", 0.0085),
                ("gammaw", 0.615),
                ("isacs", 0),
                ("ktrpn", 0.1),
                ("ku", 0.04),
                ("kuw", 0.182),
                ("kws", 0.012),
                ("mode", 1),
                ("ntm", 2.4),
                ("ntrpn", 2.0),
                ("p_a", 2.1),
                ("p_b", 9.1),
                ("p_k", 7.0),
                ("phi", 2.23),
                ("rs", 0.25),
                ("rw", 0.5),
                ("CaMKo", 0.05),
                ("KmCaM", 0.0015),
                ("KmCaMK", 0.15),
                ("aCaMK", 0.05),
                ("bCaMK", 0.00068),
                ("PKNa", 0.01833),
                ("Gncx", 0.0008),
                ("KmCaAct", 0.00015),
                ("kasymm", 12.5),
                ("kcaoff", 5000.0),
                ("kcaon", 1500000.0),
                ("kna1", 15.0),
                ("kna2", 5.0),
                ("kna3", 88.12),
                ("qca", 0.167),
                ("qna", 0.5224),
                ("wca", 60000.0),
                ("wna", 60000.0),
                ("wnaca", 5000.0),
                ("H", 1e-07),
                ("Khp", 1.698e-07),
                ("Kki", 0.5),
                ("Kko", 0.3582),
                ("Kmgatp", 1.698e-07),
                ("Knai0", 9.073),
                ("Knao0", 27.78),
                ("Knap", 224.0),
                ("Kxkur", 292.0),
                ("MgADP", 0.05),
                ("MgATP", 9.8),
                ("Pnak", 30.0),
                ("delta", -0.155),
                ("eP", 4.2),
                ("k1m", 182.4),
                ("k1p", 949.5),
                ("k2m", 39.4),
                ("k2p", 687.2),
                ("k3m", 79300.0),
                ("k3p", 1899.0),
                ("k4m", 40.0),
                ("k4p", 639.0),
                ("zk", 1.0),
                ("GKb", 0.003),
                ("PNab", 3.75e-10),
                ("PCab", 2.5e-08),
                ("GpCa", 0.0005),
                ("Esac_ns", -10.0),
                ("Gsac_k", 1.097904761904762),
                ("Gsac_ns", 0.006),
                ("lambda_max", 1.1),
                ("amp", -80.0),
                ("duration", 0.5),
                ("BSLmax", 1.124),
                ("BSRmax", 0.047),
                ("KmBSL", 0.0087),
                ("KmBSR", 0.00087),
                ("cmdnmax", 0.05),
                ("csqnmax", 10.0),
                ("kmcmdn", 0.00238),
                ("kmcsqn", 0.8),
                ("kmtrpn", 0.0005),
                ("trpnmax", 0.07),
                # Drug factors
                ("scale_drug_INa", 1.0),
                ("scale_drug_INaL", 1.0),
                ("scale_drug_Ito", 1.0),
                ("scale_drug_ICaL", 1.0),
                ("scale_drug_IKr", 1.0),
                ("scale_drug_IKs", 1.0),
                ("scale_drug_IK1", 1.0),
                ("scale_drug_IKb", 1.0),
                ("scale_drug_INab", 1.0),
                ("scale_drug_ICab", 1.0),
                ("scale_drug_IpCa", 1.0),
                ("scale_drug_Isacns", 1.0),
                ("scale_drug_Isack", 1.0),
                # Population factors
                ("scale_popu_GNa", 1.0),
                ("scale_popu_GCaL", 1.0),
                ("scale_popu_Gto", 1.0),
                ("scale_popu_GNaL", 1.0),
                ("scale_popu_GKr", 1.0),
                ("scale_popu_GKs", 1.0),
                ("scale_popu_GK1", 1.0),
                ("scale_popu_GbK", 1.0),
                ("scale_popu_KNCX", 1.0),
                ("scale_popu_KNaK", 1.0),
                ("scale_popu_GbNa", 1.0),
                ("scale_popu_GbCa", 1.0),
                ("scale_popu_KpCa", 1.0),
                ("scale_popu_KRyR", 1.0),
                ("scale_popu_KSERCA", 1.0),
                ("scale_popu_Kleak", 1.0),
                ("scale_popu_nTm", 1.0),
                ("scale_popu_CaT50ref", 1.0),
                ("scale_popu_Tref", 1.0),
                ("scale_popu_kuw", 1.0),
                ("scale_popu_kws", 1.0),
                ("scale_popu_kTRPN", 1.0),
                ("scale_popu_nTRPN", 1.0),
                ("scale_popu_ku", 1.0),
                ("scale_popu_TRPN50", 1.0),
                ("scale_popu_rw", 1.0),
                ("scale_popu_rs", 1.0),
                # HF scaling factors
                ("HF_scaling_CaMKa", 1.0),
                ("HF_scaling_Jrel_inf", 1.0),
                ("HF_scaling_Jleak", 1.0),
                ("HF_scaling_Jup", 1.0),
                ("HF_scaling_GNaL", 1.0),
                ("HF_scaling_GK1", 1.0),
                ("HF_scaling_thL", 1.0),
                ("HF_scaling_Gto", 1.0),
                ("HF_scaling_Gncx", 1.0),
                ("HF_scaling_Pnak", 1.0),
                ("HF_scaling_cat50_ref", 1.0),
            ],
        )

        if disease_state.lower() == "hf":
            logger.info("Update scaling parameters for heart failure model")
            params["HF_scaling_CaMKa"] = 1.50
            params["HF_scaling_Jrel_inf"] = pow(0.8, 8.0)
            params["HF_scaling_Jleak"] = 1.3
            params["HF_scaling_Jup"] = 0.45
            params["HF_scaling_GNaL"] = 1.3
            params["HF_scaling_GK1"] = 0.68
            params["HF_scaling_thL"] = 1.8
            params["HF_scaling_Gto"] = 0.4
            params["HF_scaling_Gncx"] = 1.6
            params["HF_scaling_Pnak"] = 0.7
            params["HF_scaling_cat50_ref"] = 0.6

        return params

    @staticmethod
    def default_initial_conditions():
        "Set-up and return default initial conditions."
        ic = OrderedDict(
            [
                ("v", -87),
                ("CaMKt", 0),
                ("m", 0),
                ("hf", 1),
                ("hs", 1),
                ("j", 1),
                ("hsp", 1),
                ("jp", 1),
                ("mL", 0),
                ("hL", 1),
                ("hLp", 1),
                ("a", 0),
                ("iF", 1),
                ("iS", 1),
                ("ap", 0),
                ("iFp", 1),
                ("iSp", 1),
                ("d", 0),
                ("ff", 1),
                ("fs", 1),
                ("fcaf", 1),
                ("fcas", 1),
                ("jca", 1),
                ("ffp", 1),
                ("fcafp", 1),
                ("nca", 0),
                ("xrf", 0),
                ("xrs", 0),
                ("xs1", 0),
                ("xs2", 0),
                ("xk1", 1),
                ("Jrelnp", 0),
                ("Jrelp", 0),
                ("nai", 7),
                ("nass", 7),
                ("ki", 145),
                ("kss", 145),
                ("cass", 0.0001),
                ("cansr", 1.2),
                ("cajsr", 1.2),
                ("XS", 0),
                ("XW", 0),
                ("CaTrpn", 0),
                ("TmB", 1),
                ("Cd", 0),
                ("cai", 0.0001),
                ("lmbda", 1),
                ("Zetas", 0),
                ("Zetaw", 0),
            ],
        )
        return ic

    def _I(self, v, s, time):
        """
        Original gotran transmembrane current dV/dt
        """
        time = time if time else Constant(0.0)

        # Assign states
        assert len(s) == 48
        (
            CaMKt,
            m,
            hf,
            hs,
            j,
            hsp,
            jp,
            mL,
            hL,
            hLp,
            a,
            iF,
            iS,
            ap,
            iFp,
            iSp,
            d,
            ff,
            fs,
            fcaf,
            fcas,
            jca,
            ffp,
            fcafp,
            nca,
            xrf,
            xrs,
            xs1,
            xs2,
            xk1,
            Jrelnp,
            Jrelp,
            nai,
            nass,
            ki,
            kss,
            cass,
            cansr,
            cajsr,
            XS,
            XW,
            CaTrpn,
            TmB,
            Cd,
            cai,
            lmbda,
            Zetas,
            Zetaw,
        ) = s

        # Assign parameters
        scale_ICaL = self._parameters["scale_ICaL"]
        scale_IK1 = self._parameters["scale_IK1"]
        scale_IKr = self._parameters["scale_IKr"]
        scale_IKs = self._parameters["scale_IKs"]
        scale_INaL = self._parameters["scale_INaL"]
        cao = self._parameters["cao"]
        ko = self._parameters["ko"]
        nao = self._parameters["nao"]
        F = self._parameters["F"]
        R = self._parameters["R"]
        T = self._parameters["T"]
        CaMKo = self._parameters["CaMKo"]
        KmCaM = self._parameters["KmCaM"]
        KmCaMK = self._parameters["KmCaMK"]
        PKNa = self._parameters["PKNa"]
        Ahf = self._parameters["Ahf"]
        GNa = self._parameters["GNa"]
        Gto = self._parameters["Gto"]
        Aff = self._parameters["Aff"]
        zca = self._parameters["zca"]
        Gncx = self._parameters["Gncx"]
        KmCaAct = self._parameters["KmCaAct"]
        kasymm = self._parameters["kasymm"]
        kcaoff = self._parameters["kcaoff"]
        kcaon = self._parameters["kcaon"]
        kna1 = self._parameters["kna1"]
        kna2 = self._parameters["kna2"]
        kna3 = self._parameters["kna3"]
        qca = self._parameters["qca"]
        qna = self._parameters["qna"]
        wca = self._parameters["wca"]
        wna = self._parameters["wna"]
        wnaca = self._parameters["wnaca"]
        H = self._parameters["H"]
        Khp = self._parameters["Khp"]
        Kki = self._parameters["Kki"]
        Kko = self._parameters["Kko"]
        Kmgatp = self._parameters["Kmgatp"]
        Knai0 = self._parameters["Knai0"]
        Knao0 = self._parameters["Knao0"]
        Knap = self._parameters["Knap"]
        Kxkur = self._parameters["Kxkur"]
        MgADP = self._parameters["MgADP"]
        MgATP = self._parameters["MgATP"]
        Pnak = self._parameters["Pnak"]
        delta = self._parameters["delta"]
        eP = self._parameters["eP"]
        k1m = self._parameters["k1m"]
        k1p = self._parameters["k1p"]
        k2m = self._parameters["k2m"]
        k2p = self._parameters["k2p"]
        k3m = self._parameters["k3m"]
        k3p = self._parameters["k3p"]
        k4m = self._parameters["k4m"]
        k4p = self._parameters["k4p"]
        zk = self._parameters["zk"]
        GKb = self._parameters["GKb"]
        PNab = self._parameters["PNab"]
        PCab = self._parameters["PCab"]
        GpCa = self._parameters["GpCa"]
        # Drug factor
        scale_drug_INa = self._parameters["scale_drug_INa"]
        scale_drug_INaL = self._parameters["scale_drug_INaL"]
        scale_drug_Ito = self._parameters["scale_drug_Ito"]
        scale_drug_ICaL = self._parameters["scale_drug_ICaL"]
        scale_drug_IKr = self._parameters["scale_drug_IKr"]
        scale_drug_IKs = self._parameters["scale_drug_IKs"]
        scale_drug_IK1 = self._parameters["scale_drug_IK1"]
        scale_drug_IKb = self._parameters["scale_drug_IKb"]
        scale_drug_INab = self._parameters["scale_drug_INab"]
        scale_drug_ICab = self._parameters["scale_drug_ICab"]
        scale_drug_IpCa = self._parameters["scale_drug_IpCa"]
        scale_drug_Isacns = self._parameters["scale_drug_Isacns"]
        scale_drug_Isack = self._parameters["scale_drug_Isack"]
        # Population factors
        scale_popu_GNa = self._parameters["scale_popu_GNa"]
        scale_popu_GCaL = self._parameters["scale_popu_GCaL"]
        scale_popu_Gto = self._parameters["scale_popu_Gto"]
        scale_popu_GNaL = self._parameters["scale_popu_GNaL"]
        scale_popu_GKr = self._parameters["scale_popu_GKr"]
        scale_popu_GKs = self._parameters["scale_popu_GKs"]
        scale_popu_GK1 = self._parameters["scale_popu_GK1"]
        scale_popu_GbK = self._parameters["scale_popu_GbK"]
        scale_popu_KNCX = self._parameters["scale_popu_KNCX"]
        scale_popu_KNaK = self._parameters["scale_popu_KNaK"]
        scale_popu_GbNa = self._parameters["scale_popu_GbNa"]
        scale_popu_GbCa = self._parameters["scale_popu_GbCa"]
        scale_popu_KpCa = self._parameters["scale_popu_KpCa"]
        # Systolic Heart Failure (HF with preserved ejection fraction)
        HF_scaling_CaMKa = self._parameters["HF_scaling_CaMKa"]
        HF_scaling_GNaL = self._parameters["HF_scaling_GNaL"]
        HF_scaling_GK1 = self._parameters["HF_scaling_GK1"]
        HF_scaling_Gto = self._parameters["HF_scaling_Gto"]
        HF_scaling_Gncx = self._parameters["HF_scaling_Gncx"]
        HF_scaling_Pnak = self._parameters["HF_scaling_Pnak"]

        # Init return args
        current = [ufl.zero()] * 1

        # Expressions for the CaMKt component
        CaMKb = CaMKo * (1.0 - CaMKt) / (1.0 + KmCaM / cass)
        CaMKa = (CaMKb + CaMKt) * HF_scaling_CaMKa

        # Expressions for the reversal potentials component
        ENa = R * T * ufl.ln(nao / nai) / F
        EK = R * T * ufl.ln(ko / ki) / F
        EKs = R * T * ufl.ln((ko + PKNa * nao) / (PKNa * nai + ki)) / F
        vffrt = (F * F) * v / (R * T)
        vfrt = F * v / (R * T)

        # Expressions for the I_Na component
        Ahs = 1.0 - Ahf
        h = Ahf * hf + Ahs * hs
        hp = Ahf * hf + Ahs * hsp
        fINap = 1.0 / (1.0 + KmCaMK / CaMKa)
        INa = (
            GNa
            * scale_drug_INa
            * scale_popu_GNa
            * ufl.elem_pow(m, 3.0)
            * (-ENa + v)
            * ((1.0 - fINap) * h * j + fINap * hp * jp)
        )

        # Expressions for the INaL component
        GNaL = 0.0075 * scale_INaL * scale_drug_INaL * scale_popu_GNaL * HF_scaling_GNaL
        fINaLp = 1.0 / (1.0 + KmCaMK / CaMKa)
        INaL = (-ENa + v) * ((1.0 - fINaLp) * hL + fINaLp * hLp) * GNaL * mL

        # Expressions for the Ito component
        AiF = 1.0 / (1.0 + 0.24348537187522867 * ufl.exp(0.006613756613756614 * v))
        AiS = 1.0 - AiF
        i = AiF * iF + AiS * iS
        ip = AiF * iFp + AiS * iSp
        fItop = 1.0 / (1.0 + KmCaMK / CaMKa)
        Ito = (
            Gto
            * scale_drug_Ito
            * scale_popu_Gto
            * HF_scaling_Gto
            * (-EK + v)
            * ((1.0 - fItop) * a * i + ap * fItop * ip)
        )

        # Expressions for the ICaL ICaNa ICaK component
        Afs = 1.0 - Aff
        f = Aff * ff + Afs * fs
        Afcaf = 0.3 + 0.6 / (1.0 + 0.36787944117144233 * ufl.exp(0.1 * v))
        Afcas = 1.0 - Afcaf
        fca = Afcaf * fcaf + Afcas * fcas
        fp = Aff * ffp + Afs * fs
        fcap = Afcaf * fcafp + Afcas * fcas
        PhiCaL = (
            4.0
            * (-0.341 * cao + cass * ufl.exp(2.0 * vfrt))
            * vffrt
            / (-1.0 + ufl.exp(2.0 * vfrt))
        )
        PhiCaNa = (
            1.0
            * (-0.75 * nao + 0.75 * ufl.exp(1.0 * vfrt) * nass)
            * vffrt
            / (-1.0 + ufl.exp(1.0 * vfrt))
        )
        PhiCaK = (
            1.0
            * (-0.75 * ko + 0.75 * ufl.exp(1.0 * vfrt) * kss)
            * vffrt
            / (-1.0 + ufl.exp(1.0 * vfrt))
        )
        PCa = 0.0001 * scale_ICaL * scale_drug_ICaL * scale_popu_GCaL
        PCap = 1.1 * PCa
        PCaNa = 0.00125 * PCa
        PCaK = 0.0003574 * PCa
        PCaNap = 0.00125 * PCap
        PCaKp = 0.0003574 * PCap
        fICaLp = 1.0 / (1.0 + KmCaMK / CaMKa)
        ICaL = (1.0 - fICaLp) * (
            (1.0 - nca) * f + fca * jca * nca
        ) * PCa * PhiCaL * d + (
            (1.0 - nca) * fp + fcap * jca * nca
        ) * PCap * PhiCaL * d * fICaLp
        ICaNa = (1.0 - fICaLp) * (
            (1.0 - nca) * f + fca * jca * nca
        ) * PCaNa * PhiCaNa * d + (
            (1.0 - nca) * fp + fcap * jca * nca
        ) * PCaNap * PhiCaNa * d * fICaLp
        ICaK = (1.0 - fICaLp) * (
            (1.0 - nca) * f + fca * jca * nca
        ) * PCaK * PhiCaK * d + (
            (1.0 - nca) * fp + fcap * jca * nca
        ) * PCaKp * PhiCaK * d * fICaLp

        # Expressions for the IKr component
        Axrf = 1.0 / (1.0 + 4.197299094734718 * ufl.exp(0.02617115938236064 * v))
        Axrs = 1.0 - Axrf
        xr = Axrf * xrf + Axrs * xrs
        rkr = 1.0 / (
            (1.0 + 2.0820090840784555 * ufl.exp(0.013333333333333334 * v))
            * (1.0 + 0.7165313105737893 * ufl.exp(0.03333333333333333 * v))
        )
        GKr = 0.046 * scale_IKr * scale_drug_IKr * scale_popu_GKr
        IKr = 0.4303314829119352 * ufl.sqrt(ko) * (-EK + v) * GKr * rkr * xr

        # Expressions for the IKs component
        KsCa = 1.0 + 0.6 / (1.0 + 6.481821026062645e-07 * ufl.elem_pow(1.0 / cai, 1.4))
        GKs = 0.0034 * scale_IKs * scale_drug_IKs * scale_popu_GKs
        IKs = (-EKs + v) * GKs * KsCa * xs1 * xs2
        rk1 = 1.0 / (
            1.0
            + 69220.6322106767
            * ufl.exp(0.10534077741493732 * v - 0.27388602127883704 * ko)
        )
        GK1 = 0.1908 * scale_IK1 * scale_drug_IK1 * scale_popu_GK1 * HF_scaling_GK1
        IK1 = ufl.sqrt(ko) * (-EK + v) * GK1 * rk1 * xk1

        # Expressions for the INaCa_i component
        hca = ufl.exp(F * qca * v / (R * T))
        hna = ufl.exp(F * qna * v / (R * T))
        h1_i = 1 + (1 + hna) * nai / kna3
        h2_i = hna * nai / (kna3 * h1_i)
        h3_i = 1.0 / h1_i
        h4_i = 1.0 + (1 + nai / kna2) * nai / kna1
        h5_i = (nai * nai) / (kna1 * kna2 * h4_i)
        h6_i = 1.0 / h4_i
        h7_i = 1.0 + nao * (1.0 + 1.0 / hna) / kna3
        h8_i = nao / (kna3 * h7_i * hna)
        h9_i = 1.0 / h7_i
        h10_i = 1.0 + kasymm + nao * (1.0 + nao / kna2) / kna1
        h11_i = (nao * nao) / (kna1 * kna2 * h10_i)
        h12_i = 1.0 / h10_i
        k1_i = cao * kcaon * h12_i
        k2_i = kcaoff
        k3p_i = wca * h9_i
        k3pp_i = wnaca * h8_i
        k3_i = k3p_i + k3pp_i
        k4p_i = wca * h3_i / hca
        k4pp_i = wnaca * h2_i
        k4_i = k4p_i + k4pp_i
        k5_i = kcaoff
        k6_i = kcaon * cai * h6_i
        k7_i = wna * h2_i * h5_i
        k8_i = wna * h11_i * h8_i
        x1_i = (k2_i + k3_i) * k5_i * k7_i + (k6_i + k7_i) * k2_i * k4_i
        x2_i = (k1_i + k8_i) * k4_i * k6_i + (k4_i + k5_i) * k1_i * k7_i
        x3_i = (k2_i + k3_i) * k6_i * k8_i + (k6_i + k7_i) * k1_i * k3_i
        x4_i = (k1_i + k8_i) * k3_i * k5_i + (k4_i + k5_i) * k2_i * k8_i
        E1_i = x1_i / (x1_i + x2_i + x3_i + x4_i)
        E2_i = x2_i / (x1_i + x2_i + x3_i + x4_i)
        E3_i = x3_i / (x1_i + x2_i + x3_i + x4_i)
        E4_i = x4_i / (x1_i + x2_i + x3_i + x4_i)
        allo_i = 1.0 / (1.0 + ufl.elem_pow(KmCaAct / cai, 2.0))
        zna = 1.0
        JncxNa_i = E3_i * k4pp_i - E2_i * k3pp_i + 3.0 * E4_i * k7_i - 3.0 * E1_i * k8_i
        JncxCa_i = E2_i * k2_i - E1_i * k1_i
        INaCa_i = (
            0.8
            * Gncx
            * scale_popu_KNCX
            * HF_scaling_Gncx
            * (zca * JncxCa_i + zna * JncxNa_i)
            * allo_i
        )

        # Expressions for the INaCa_ss component
        h1 = 1.0 + (1.0 + hna) * nass / kna3
        h2 = hna * nass / (kna3 * h1)
        h3 = 1.0 / h1
        h4 = 1.0 + (1 + nass / kna2) * nass / kna1
        h5 = (nass * nass) / (kna1 * kna2 * h4)
        h6 = 1.0 / h4
        h7 = 1.0 + nao * (1.0 + 1.0 / hna) / kna3
        h8 = nao / (kna3 * h7 * hna)
        h9 = 1.0 / h7
        h10 = 1.0 + kasymm + nao * (1 + nao / kna2) / kna1
        h11 = (nao * nao) / (kna1 * kna2 * h10)
        h12 = 1.0 / h10
        k1 = cao * kcaon * h12
        k2 = kcaoff
        k3p_ss = wca * h9
        k3pp = wnaca * h8
        k3 = k3p_ss + k3pp
        k4p_ss = wca * h3 / hca
        k4pp = wnaca * h2
        k4 = k4p_ss + k4pp
        k5 = kcaoff
        k6 = kcaon * cass * h6
        k7 = wna * h2 * h5
        k8 = wna * h11 * h8
        x1_ss = (k2 + k3) * k5 * k7 + (k6 + k7) * k2 * k4
        x2_ss = (k1 + k8) * k4 * k6 + (k4 + k5) * k1 * k7
        x3_ss = (k2 + k3) * k6 * k8 + (k6 + k7) * k1 * k3
        x4_ss = (k1 + k8) * k3 * k5 + (k4 + k5) * k2 * k8
        E1_ss = x1_ss / (x1_ss + x2_ss + x3_ss + x4_ss)
        E2_ss = x2_ss / (x1_ss + x2_ss + x3_ss + x4_ss)
        E3_ss = x3_ss / (x1_ss + x2_ss + x3_ss + x4_ss)
        E4_ss = x4_ss / (x1_ss + x2_ss + x3_ss + x4_ss)
        allo_ss = 1.0 / (1.0 + ufl.elem_pow(KmCaAct / cass, 2.0))
        JncxNa_ss = E3_ss * k4pp - E2_ss * k3pp + 3.0 * E4_ss * k7 - 3.0 * E1_ss * k8
        JncxCa_ss = E2_ss * k2 - E1_ss * k1
        INaCa_ss = (
            0.2
            * Gncx
            * scale_popu_KNCX
            * HF_scaling_Gncx
            * (zca * JncxCa_ss + zna * JncxNa_ss)
            * allo_ss
        )

        # Expressions for the INaK component
        Knai = Knai0 * ufl.exp(0.3333333333333333 * F * delta * v / (R * T))
        Knao = Knao0 * ufl.exp(0.3333333333333333 * F * (1.0 - delta) * v / (R * T))
        P = eP / (1.0 + H / Khp + nai / Knap + ki / Kxkur)
        a1 = (
            k1p
            * ufl.elem_pow(nai / Knai, 3.0)
            / (
                -1.0
                + ufl.elem_pow(1.0 + ki / Kki, 2.0)
                + ufl.elem_pow(1.0 + nai / Knai, 3.0)
            )
        )
        b1 = MgADP * k1m
        a2 = k2p
        b2 = (
            k2m
            * ufl.elem_pow(nao / Knao, 3.0)
            / (
                -1.0
                + ufl.elem_pow(1.0 + ko / Kko, 2.0)
                + ufl.elem_pow(1.0 + nao / Knao, 3.0)
            )
        )
        a3 = (
            k3p
            * ufl.elem_pow(ko / Kko, 2.0)
            / (
                -1.0
                + ufl.elem_pow(1.0 + ko / Kko, 2.0)
                + ufl.elem_pow(1.0 + nao / Knao, 3.0)
            )
        )
        b3 = H * k3m * P / (1.0 + MgATP / Kmgatp)
        a4 = MgATP * k4p / (Kmgatp * (1.0 + MgATP / Kmgatp))
        b4 = (
            k4m
            * ufl.elem_pow(ki / Kki, 2.0)
            / (
                -1.0
                + ufl.elem_pow(1.0 + ki / Kki, 2.0)
                + ufl.elem_pow(1.0 + nai / Knai, 3.0)
            )
        )
        x1 = a1 * a2 * a4 + a1 * a2 * b3 + a2 * b3 * b4 + b2 * b3 * b4
        x2 = a1 * a2 * a3 + a2 * a3 * b4 + a3 * b1 * b4 + b1 * b2 * b4
        x3 = a2 * a3 * a4 + a3 * a4 * b1 + a4 * b1 * b2 + b1 * b2 * b3
        x4 = a1 * a3 * a4 + a1 * a4 * b2 + a1 * b2 * b3 + b2 * b3 * b4
        E1 = x1 / (x1 + x2 + x3 + x4)
        E2 = x2 / (x1 + x2 + x3 + x4)
        E3 = x3 / (x1 + x2 + x3 + x4)
        E4 = x4 / (x1 + x2 + x3 + x4)
        JnakNa = 3.0 * E1 * a3 - 3.0 * E2 * b3
        JnakK = 2.0 * E4 * b1 - 2.0 * E3 * a1
        INaK = Pnak * scale_popu_KNaK * HF_scaling_Pnak * (zk * JnakK + zna * JnakNa)

        # Expressions for the IKb component
        xkb = 1.0 / (1.0 + 2.202363450949239 * ufl.exp(-0.05452562704471101 * v))
        IKb = GKb * scale_popu_GbK * scale_drug_IKb * (-EK + v) * xkb

        # Expressions for the INab component
        INab = (
            PNab
            * scale_drug_INab
            * scale_popu_GbNa
            * (-nao + ufl.exp(vfrt) * nai)
            * vffrt
            / (-1.0 + ufl.exp(vfrt))
        )

        # Expressions for the ICab component
        ICab = (
            4.0
            * PCab
            * scale_drug_ICab
            * scale_popu_GbCa
            * (-0.341 * cao + cai * ufl.exp(2.0 * vfrt))
            * vffrt
            / (-1.0 + ufl.exp(2.0 * vfrt))
        )

        # Expressions for the IpCa component
        IpCa = GpCa * scale_drug_IpCa * scale_popu_KpCa * cai / (0.0005 + cai)

        # Expressions for the Isac (Pueyo)--> ns + k component
        Gsac_ns = 0.006 * scale_drug_Isacns
        Esac_ns = -10.0
        lambda_max = 1.1
        Gsac_k = (0.2882 * 800.0 / 210.0) * scale_drug_Isack  # Pueyo endo

        Isac_P_ns = ufl.conditional(
            ufl.lt(lmbda, 1.0),
            0.0,
            Gsac_ns * ((lmbda - 1.0) / (lambda_max - 1.0)) * (v - Esac_ns),
        )
        Isac_P_k = ufl.conditional(
            ufl.lt(lmbda, 1.0),
            0.0,
            Gsac_k
            * ((lmbda - 1.0) / (lambda_max - 1.0))
            * (1.0 / (1.0 + ufl.exp((19.05 - v) / (29.98)))),
        )

        # Expressions for the Istim component
        Istim = 0  # amp*(ufl.le(time, duration))

        # Expressions for the membrane potential component
        current[0] = (
            -Isac_P_k
            - Isac_P_ns
            - ICaK
            - ICaL
            - ICaNa
            - ICab
            - IK1
            - IKb
            - IKr
            - IKs
            - INa
            - INaCa_i
            - INaCa_ss
            - INaK
            - INaL
            - INab
            - IpCa
            - Istim
            - Ito
        )

        # Return results
        return current[0]

    def I(self, v, s, time=None):  # noqa: E741, E743
        """
        Transmembrane current

           I = -dV/dt

        """
        return -self._I(v, s, time)

    def F(self, v, s, time=None):
        """
        Right hand side for ODE system
        """
        time = time if time else Constant(0.0)

        # Assign states
        assert len(s) == 48
        (
            CaMKt,
            m,
            hf,
            hs,
            j,
            hsp,
            jp,
            mL,
            hL,
            hLp,
            a,
            iF,
            iS,
            ap,
            iFp,
            iSp,
            d,
            ff,
            fs,
            fcaf,
            fcas,
            jca,
            ffp,
            fcafp,
            nca,
            xrf,
            xrs,
            xs1,
            xs2,
            xk1,
            Jrelnp,
            Jrelp,
            nai,
            nass,
            ki,
            kss,
            cass,
            cansr,
            cajsr,
            XS,
            XW,
            CaTrpn,
            TmB,
            Cd,
            cai,
            lmbda,
            Zetas,
            Zetaw,
        ) = s

        # Assign parameters
        scale_ICaL = self._parameters["scale_ICaL"]
        scale_IK1 = self._parameters["scale_IK1"]
        scale_IKr = self._parameters["scale_IKr"]
        scale_IKs = self._parameters["scale_IKs"]
        scale_INaL = self._parameters["scale_INaL"]
        cao = self._parameters["cao"]
        ko = self._parameters["ko"]
        nao = self._parameters["nao"]
        F = self._parameters["F"]
        R = self._parameters["R"]
        T = self._parameters["T"]
        L = self._parameters["L"]
        rad = self._parameters["rad"]
        CaMKo = self._parameters["CaMKo"]
        KmCaM = self._parameters["KmCaM"]
        KmCaMK = self._parameters["KmCaMK"]
        aCaMK = self._parameters["aCaMK"]
        bCaMK = self._parameters["bCaMK"]
        PKNa = self._parameters["PKNa"]
        Ahf = self._parameters["Ahf"]
        GNa = self._parameters["GNa"]
        thL = self._parameters["thL"]
        Gto = self._parameters["Gto"]
        delta_epi = self._parameters["delta_epi"]
        Aff = self._parameters["Aff"]
        Kmn = self._parameters["Kmn"]
        k2n = self._parameters["k2n"]
        tjca = self._parameters["tjca"]
        zca = self._parameters["zca"]
        Gncx = self._parameters["Gncx"]
        KmCaAct = self._parameters["KmCaAct"]
        kasymm = self._parameters["kasymm"]
        kcaoff = self._parameters["kcaoff"]
        kcaon = self._parameters["kcaon"]
        kna1 = self._parameters["kna1"]
        kna2 = self._parameters["kna2"]
        kna3 = self._parameters["kna3"]
        qca = self._parameters["qca"]
        qna = self._parameters["qna"]
        wca = self._parameters["wca"]
        wna = self._parameters["wna"]
        wnaca = self._parameters["wnaca"]
        H = self._parameters["H"]
        Khp = self._parameters["Khp"]
        Kki = self._parameters["Kki"]
        Kko = self._parameters["Kko"]
        Kmgatp = self._parameters["Kmgatp"]
        Knai0 = self._parameters["Knai0"]
        Knao0 = self._parameters["Knao0"]
        Knap = self._parameters["Knap"]
        Kxkur = self._parameters["Kxkur"]
        MgADP = self._parameters["MgADP"]
        MgATP = self._parameters["MgATP"]
        Pnak = self._parameters["Pnak"]
        delta = self._parameters["delta"]
        eP = self._parameters["eP"]
        k1m = self._parameters["k1m"]
        k1p = self._parameters["k1p"]
        k2m = self._parameters["k2m"]
        k2p = self._parameters["k2p"]
        k3m = self._parameters["k3m"]
        k3p = self._parameters["k3p"]
        k4m = self._parameters["k4m"]
        k4p = self._parameters["k4p"]
        zk = self._parameters["zk"]
        GKb = self._parameters["GKb"]
        PNab = self._parameters["PNab"]
        PCab = self._parameters["PCab"]
        GpCa = self._parameters["GpCa"]
        bt = self._parameters["bt"]
        BSLmax = self._parameters["BSLmax"]
        BSRmax = self._parameters["BSRmax"]
        KmBSL = self._parameters["KmBSL"]
        KmBSR = self._parameters["KmBSR"]
        cmdnmax = self._parameters["cmdnmax"]
        csqnmax = self._parameters["csqnmax"]
        kmcmdn = self._parameters["kmcmdn"]
        kmcsqn = self._parameters["kmcsqn"]
        trpnmax = self._parameters["trpnmax"]
        Beta1 = self._parameters["Beta1"]

        Trpn50 = self._parameters["Trpn50"]
        cat50_ref = self._parameters["cat50_ref"]

        etal = self._parameters["etal"]
        etas = self._parameters["etas"]
        gammas = self._parameters["gammas"]
        gammaw = self._parameters["gammaw"]
        ktrpn = self._parameters["ktrpn"]
        ku = self._parameters["ku"]
        kuw = self._parameters["kuw"]
        kws = self._parameters["kws"]
        # lmbda = self._parameters["lmbda"]
        ntm = self._parameters["ntm"]
        ntrpn = self._parameters["ntrpn"]
        p_k = self._parameters["p_k"]

        rs = self._parameters["rs"]
        rw = self._parameters["rw"]

        # Zetas = self._parameters["Zetas"]
        # Zetaw = self._parameters["Zetaw"]

        # Drug factor
        scale_drug_INa = self._parameters["scale_drug_INa"]
        scale_drug_INaL = self._parameters["scale_drug_INaL"]
        scale_drug_Ito = self._parameters["scale_drug_Ito"]
        scale_drug_ICaL = self._parameters["scale_drug_ICaL"]
        scale_drug_IKr = self._parameters["scale_drug_IKr"]
        scale_drug_IKs = self._parameters["scale_drug_IKs"]
        scale_drug_IK1 = self._parameters["scale_drug_IK1"]
        scale_drug_IKb = self._parameters["scale_drug_IKb"]
        scale_drug_INab = self._parameters["scale_drug_INab"]
        scale_drug_ICab = self._parameters["scale_drug_ICab"]
        scale_drug_IpCa = self._parameters["scale_drug_IpCa"]
        scale_drug_Isacns = self._parameters["scale_drug_Isacns"]
        scale_drug_Isack = self._parameters["scale_drug_Isack"]

        # Population factors
        scale_popu_GNa = self._parameters["scale_popu_GNa"]
        scale_popu_GCaL = self._parameters["scale_popu_GCaL"]
        scale_popu_Gto = self._parameters["scale_popu_Gto"]
        scale_popu_GNaL = self._parameters["scale_popu_GNaL"]
        scale_popu_GKr = self._parameters["scale_popu_GKr"]
        scale_popu_GKs = self._parameters["scale_popu_GKs"]
        scale_popu_GK1 = self._parameters["scale_popu_GK1"]
        scale_popu_GbK = self._parameters["scale_popu_GbK"]
        scale_popu_KNCX = self._parameters["scale_popu_KNCX"]
        scale_popu_KNaK = self._parameters["scale_popu_KNaK"]
        scale_popu_GbNa = self._parameters["scale_popu_GbNa"]
        scale_popu_GbCa = self._parameters["scale_popu_GbCa"]
        scale_popu_KpCa = self._parameters["scale_popu_KpCa"]
        scale_popu_KRyR = self._parameters["scale_popu_KRyR"]
        scale_popu_KSERCA = self._parameters["scale_popu_KSERCA"]
        scale_popu_Kleak = self._parameters["scale_popu_Kleak"]
        scale_popu_nTm = self._parameters["scale_popu_nTm"]
        scale_popu_CaT50ref = self._parameters["scale_popu_CaT50ref"]
        scale_popu_kuw = self._parameters["scale_popu_kuw"]
        scale_popu_kws = self._parameters["scale_popu_kws"]
        scale_popu_kTRPN = self._parameters["scale_popu_kTRPN"]
        scale_popu_nTRPN = self._parameters["scale_popu_nTRPN"]
        scale_popu_ku = self._parameters["scale_popu_ku"]
        scale_popu_TRPN50 = self._parameters["scale_popu_TRPN50"]
        scale_popu_rw = self._parameters["scale_popu_rw"]
        scale_popu_rs = self._parameters["scale_popu_rs"]

        # Systolic Heart Failure (HF with preserved ejection fraction)
        HF_scaling_CaMKa = self._parameters["HF_scaling_CaMKa"]
        HF_scaling_Jrel_inf = self._parameters["HF_scaling_Jrel_inf"]
        HF_scaling_Jleak = self._parameters["HF_scaling_Jleak"]
        HF_scaling_Jup = self._parameters["HF_scaling_Jup"]
        HF_scaling_GNaL = self._parameters["HF_scaling_GNaL"]
        HF_scaling_GK1 = self._parameters["HF_scaling_GK1"]
        HF_scaling_thL = self._parameters["HF_scaling_thL"]
        HF_scaling_Gto = self._parameters["HF_scaling_Gto"]
        HF_scaling_Gncx = self._parameters["HF_scaling_Gncx"]
        HF_scaling_Pnak = self._parameters["HF_scaling_Pnak"]
        HF_scaling_cat50_ref = self._parameters["HF_scaling_cat50_ref"]

        # Init return args
        F_expressions = [dolfin.Constant(0.0)] * 48

        # Expressions for the cell geometry component
        vcell = 3140.0 * L * (rad * rad)
        Ageo = 6.28 * (rad * rad) + 6.28 * L * rad
        Acap = 2 * Ageo
        vmyo = 0.68 * vcell
        vnsr = 0.0552 * vcell
        vjsr = 0.0048 * vcell
        vss = 0.02 * vcell

        # Expressions for the CaMKt component
        CaMKb = CaMKo * (1.0 - CaMKt) / (1.0 + KmCaM / cass)
        CaMKa = (CaMKb + CaMKt) * HF_scaling_CaMKa
        F_expressions[0] = -bCaMK * CaMKt + aCaMK * (CaMKb + CaMKt) * CaMKb

        # Expressions for the reversal potentials component
        ENa = R * T * ufl.ln(nao / nai) / F
        EK = R * T * ufl.ln(ko / ki) / F
        EKs = R * T * ufl.ln((ko + PKNa * nao) / (PKNa * nai + ki)) / F
        vffrt = (F * F) * v / (R * T)
        vfrt = F * v / (R * T)

        # Expressions for the I_Na component
        mss = 1.0 / (1.0 + 0.0014599788446489682 * ufl.exp(-0.13333333333333333 * v))
        tm = 1.0 / (
            9.454904638564724 * ufl.exp(0.02876042565429968 * v)
            + 1.9314113558536928e-05 * ufl.exp(-0.16792611251049538 * v)
        )
        F_expressions[1] = (-m + mss) / tm
        hss = 1.0 / (1 + 302724.605401998 * ufl.exp(0.1607717041800643 * v))
        thf = 1.0 / (
            1.183856958289087e-05 * ufl.exp(-0.15910898965791567 * v)
            + 6.305549185817275 * ufl.exp(0.0493339911198816 * v)
        )
        ths = 1.0 / (
            0.005164670235381792 * ufl.exp(-0.035650623885918005 * v)
            + 0.36987619372096325 * ufl.exp(0.017649135192375574 * v)
        )
        Ahs = 1.0 - Ahf
        F_expressions[2] = (-hf + hss) / thf
        F_expressions[3] = (-hs + hss) / ths
        h = Ahf * hf + Ahs * hs
        jss = hss
        tj = 2.038 + 1.0 / (
            0.3131936394738773 * ufl.exp(0.02600780234070221 * v)
            + 1.1315282095590072e-07 * ufl.exp(-0.12075836251660427 * v)
        )
        F_expressions[4] = (-j + jss) / tj
        hssp = 1.0 / (1 + 820249.0921708513 * ufl.exp(0.1607717041800643 * v))
        thsp = 3.0 * ths
        F_expressions[5] = (-hsp + hssp) / thsp
        hp = Ahf * hf + Ahs * hsp
        tjp = 1.46 * tj
        F_expressions[6] = (-jp + jss) / tjp
        fINap = 1.0 / (1.0 + KmCaMK / CaMKa)
        INa = (
            GNa
            * scale_drug_INa
            * scale_popu_GNa
            * ufl.elem_pow(m, 3.0)
            * (-ENa + v)
            * ((1.0 - fINap) * h * j + fINap * hp * jp)
        )

        # Expressions for the INaL component
        mLss = 1.0 / (1.0 + 0.000291579585635531 * ufl.exp(-0.18996960486322187 * v))
        tmL = tm
        F_expressions[7] = (-mL + mLss) / tmL
        hLss = 1.0 / (1.0 + 120578.15595522427 * ufl.exp(0.13354700854700854 * v))
        F_expressions[8] = (-hL + hLss) / (thL * HF_scaling_thL)
        hLssp = 1.0 / (1.0 + 275969.2903869871 * ufl.exp(0.13354700854700854 * v))
        thLp = 3.0 * thL * HF_scaling_thL
        F_expressions[9] = (-hLp + hLssp) / thLp
        GNaL = 0.0075 * scale_INaL * scale_drug_INaL * scale_popu_GNaL * HF_scaling_GNaL
        fINaLp = 1.0 / (1.0 + KmCaMK / CaMKa)
        INaL = (-ENa + v) * ((1.0 - fINaLp) * hL + fINaLp * hLp) * GNaL * mL

        # Expressions for the Ito component
        ass = 1.0 / (1.0 + 2.6316508161673635 * ufl.exp(-0.06747638326585695 * v))
        ta = 1.0515 / (
            1.0 / (1.2089 + 2.2621017070578837 * ufl.exp(-0.03403513787634354 * v))
            + 3.5 / (1.0 + 30.069572727397507 * ufl.exp(0.03403513787634354 * v))
        )
        F_expressions[10] = (-a + ass) / ta
        iss = 1.0 / (1.0 + 2194.970764538301 * ufl.exp(0.17510068289266328 * v))
        tiF = 4.562 + delta_epi / (
            0.14468698421272827 * ufl.exp(-0.01 * v)
            + 1.6300896349780942 * ufl.exp(0.06027727546714889 * v)
        )
        tiS = 23.62 + delta_epi / (
            0.00027617763953377436 * ufl.exp(-0.01693480101608806 * v)
            + 0.024208962804604526 * ufl.exp(0.12377769525931426 * v)
        )
        AiF = 1.0 / (1.0 + 0.24348537187522867 * ufl.exp(0.006613756613756614 * v))
        AiS = 1.0 - AiF
        F_expressions[11] = (-iF + iss) / tiF
        F_expressions[12] = (-iS + iss) / tiS
        i = AiF * iF + AiS * iS
        assp = 1.0 / (1.0 + 5.167428462230666 * ufl.exp(-0.06747638326585695 * v))
        F_expressions[13] = (-ap + assp) / ta
        dti_develop = 1.354 + 0.0001 / (
            2.6591269045230603e-05 * ufl.exp(0.06293266205160478 * v)
            + 4.5541779737128264e24 * ufl.exp(-4.642525533890436 * v)
        )
        dti_recover = 1.0 - 0.5 / (1.0 + 33.11545195869231 * ufl.exp(0.05 * v))
        tiFp = dti_develop * dti_recover * tiF
        tiSp = dti_develop * dti_recover * tiS
        F_expressions[14] = (-iFp + iss) / tiFp
        F_expressions[15] = (-iSp + iss) / tiSp
        ip = AiF * iFp + AiS * iSp
        fItop = 1.0 / (1.0 + KmCaMK / CaMKa)
        Ito = (
            Gto
            * scale_drug_Ito
            * scale_popu_Gto
            * HF_scaling_Gto
            * (-EK + v)
            * ((1.0 - fItop) * a * i + ap * fItop * ip)
        )

        # Expressions for the ICaL ICaNa ICaK component
        dss = 1.0 / (1.0 + 0.39398514226669484 * ufl.exp(-0.23640661938534277 * v))
        td = 0.6 + 1.0 / (
            3.5254214873653824 * ufl.exp(0.09 * v)
            + 0.7408182206817179 * ufl.exp(-0.05 * v)
        )
        F_expressions[16] = (-d + dss) / td
        fss = 1.0 / (1.0 + 199.86038496778565 * ufl.exp(0.27056277056277056 * v))
        tff = 7.0 + 1.0 / (
            0.03325075244518792 * ufl.exp(0.1 * v)
            + 0.0006090087745647571 * ufl.exp(-0.1 * v)
        )
        tfs = 1000.0 + 1.0 / (
            1.0027667890106652e-05 * ufl.exp(-0.25 * v)
            + 8.053415618124885e-05 * ufl.exp(0.16666666666666666 * v)
        )
        Afs = 1.0 - Aff
        F_expressions[17] = (-ff + fss) / tff
        F_expressions[18] = (-fs + fss) / tfs
        f = Aff * ff + Afs * fs
        fcass = fss
        tfcaf = 7.0 + 1.0 / (
            0.0708317980974062 * ufl.exp(-0.14285714285714285 * v)
            + 0.02258872488031037 * ufl.exp(0.14285714285714285 * v)
        )
        tfcas = 100.0 + 1.0 / (
            0.00012 * ufl.exp(0.14285714285714285 * v)
            + 0.00012 * ufl.exp(-0.3333333333333333 * v)
        )
        Afcaf = 0.3 + 0.6 / (1.0 + 0.36787944117144233 * ufl.exp(0.1 * v))
        Afcas = 1.0 - Afcaf
        F_expressions[19] = (-fcaf + fcass) / tfcaf
        F_expressions[20] = (-fcas + fcass) / tfcas
        fca = Afcaf * fcaf + Afcas * fcas
        F_expressions[21] = (-jca + fcass) / tjca
        tffp = 2.5 * tff
        F_expressions[22] = (-ffp + fss) / tffp
        fp = Aff * ffp + Afs * fs
        tfcafp = 2.5 * tfcaf
        F_expressions[23] = (-fcafp + fcass) / tfcafp
        fcap = Afcaf * fcafp + Afcas * fcas
        km2n = 1.0 * jca
        anca = 1.0 / (ufl.elem_pow(1.0 + Kmn / cass, 4.0) + k2n / km2n)
        F_expressions[24] = k2n * anca - km2n * nca
        PhiCaL = (
            4.0
            * (-0.341 * cao + cass * ufl.exp(2.0 * vfrt))
            * vffrt
            / (-1.0 + ufl.exp(2.0 * vfrt))
        )
        PhiCaNa = (
            1.0
            * (-0.75 * nao + 0.75 * ufl.exp(1.0 * vfrt) * nass)
            * vffrt
            / (-1.0 + ufl.exp(1.0 * vfrt))
        )
        PhiCaK = (
            1.0
            * (-0.75 * ko + 0.75 * ufl.exp(1.0 * vfrt) * kss)
            * vffrt
            / (-1.0 + ufl.exp(1.0 * vfrt))
        )
        PCa = 0.0001 * scale_ICaL * scale_drug_ICaL * scale_popu_GCaL
        PCap = 1.1 * PCa
        PCaNa = 0.00125 * PCa
        PCaK = 0.0003574 * PCa
        PCaNap = 0.00125 * PCap
        PCaKp = 0.0003574 * PCap
        fICaLp = 1.0 / (1.0 + KmCaMK / CaMKa)
        ICaL = (1.0 - fICaLp) * (
            (1.0 - nca) * f + fca * jca * nca
        ) * PCa * PhiCaL * d + (
            (1.0 - nca) * fp + fcap * jca * nca
        ) * PCap * PhiCaL * d * fICaLp
        ICaNa = (1.0 - fICaLp) * (
            (1.0 - nca) * f + fca * jca * nca
        ) * PCaNa * PhiCaNa * d + (
            (1.0 - nca) * fp + fcap * jca * nca
        ) * PCaNap * PhiCaNa * d * fICaLp
        ICaK = (1.0 - fICaLp) * (
            (1.0 - nca) * f + fca * jca * nca
        ) * PCaK * PhiCaK * d + (
            (1.0 - nca) * fp + fcap * jca * nca
        ) * PCaKp * PhiCaK * d * fICaLp

        # Expressions for the IKr component
        xrss = 1.0 / (1.0 + 0.29287308872377504 * ufl.exp(-0.14729709824716453 * v))
        txrf = 12.98 + 1.0 / (
            0.0001020239312894894 * ufl.exp(0.25846471956577927 * v)
            + 0.00042992960891929087 * ufl.exp(-0.04906771344455348 * v)
        )
        txrs = 1.865 + 1.0 / (
            0.0005922420036809394 * ufl.exp(0.13596193065941536 * v)
            + 3.549966111802463e-05 * ufl.exp(-0.03855050115651503 * v)
        )
        Axrf = 1.0 / (1.0 + 4.197299094734718 * ufl.exp(0.02617115938236064 * v))
        Axrs = 1.0 - Axrf
        F_expressions[25] = (-xrf + xrss) / txrf
        F_expressions[26] = (-xrs + xrss) / txrs
        xr = Axrf * xrf + Axrs * xrs
        rkr = 1.0 / (
            (1.0 + 2.0820090840784555 * ufl.exp(0.013333333333333334 * v))
            * (1.0 + 0.7165313105737893 * ufl.exp(0.03333333333333333 * v))
        )
        GKr = 0.046 * scale_IKr * scale_drug_IKr * scale_popu_GKr
        IKr = 0.4303314829119352 * ufl.sqrt(ko) * (-EK + v) * GKr * rkr * xr

        # Expressions for the IKs component
        xs1ss = 1.0 / (1.0 + 0.27288596035656526 * ufl.exp(-0.11195700850873264 * v))
        txs1 = 817.3 + 1.0 / (
            0.003504067763074858 * ufl.exp(0.056179775280898875 * v)
            + 0.0005184809083581659 * ufl.exp(-0.004347826086956522 * v)
        )
        F_expressions[27] = (-xs1 + xs1ss) / txs1
        xs2ss = xs1ss
        txs2 = 1.0 / (
            0.0022561357010639103 * ufl.exp(-0.03225806451612903 * v)
            + 0.0008208499862389881 * ufl.exp(0.05 * v)
        )
        F_expressions[28] = (-xs2 + xs2ss) / txs2
        KsCa = 1.0 + 0.6 / (1.0 + 6.481821026062645e-07 * ufl.elem_pow(1.0 / cai, 1.4))
        GKs = 0.0034 * scale_IKs * scale_drug_IKs * scale_popu_GKs
        IKs = (-EKs + v) * GKs * KsCa * xs1 * xs2
        xk1ss = 1.0 / (
            1.0 + ufl.exp((-144.59 - v - 2.5538 * ko) / (3.8115 + 1.5692 * ko))
        )
        txk1 = 122.2 / (
            0.0019352007631390235 * ufl.exp(-0.049115913555992145 * v)
            + 30.43364757524903 * ufl.exp(0.014423770373575654 * v)
        )
        F_expressions[29] = (-xk1 + xk1ss) / txk1
        rk1 = 1.0 / (
            1.0
            + 69220.6322106767
            * ufl.exp(0.10534077741493732 * v - 0.27388602127883704 * ko)
        )
        GK1 = 0.1908 * scale_IK1 * scale_drug_IK1 * scale_popu_GK1 * HF_scaling_GK1
        IK1 = ufl.sqrt(ko) * (-EK + v) * GK1 * rk1 * xk1

        # Expressions for the INaCa_i component
        hca = ufl.exp(F * qca * v / (R * T))
        hna = ufl.exp(F * qna * v / (R * T))
        h1_i = 1.0 + (1.0 + hna) * nai / kna3
        h2_i = hna * nai / (kna3 * h1_i)
        h3_i = 1.0 / h1_i
        h4_i = 1.0 + (1 + nai / kna2) * nai / kna1
        h5_i = (nai * nai) / (kna1 * kna2 * h4_i)
        h6_i = 1.0 / h4_i
        h7_i = 1.0 + nao * (1.0 + 1.0 / hna) / kna3
        h8_i = nao / (kna3 * h7_i * hna)
        h9_i = 1.0 / h7_i
        h10_i = 1.0 + kasymm + nao * (1.0 + nao / kna2) / kna1
        h11_i = (nao * nao) / (kna1 * kna2 * h10_i)
        h12_i = 1.0 / h10_i
        k1_i = cao * kcaon * h12_i
        k2_i = kcaoff
        k3p_i = wca * h9_i
        k3pp_i = wnaca * h8_i
        k3_i = k3p_i + k3pp_i
        k4p_i = wca * h3_i / hca
        k4pp_i = wnaca * h2_i
        k4_i = k4p_i + k4pp_i
        k5_i = kcaoff
        k6_i = kcaon * cai * h6_i
        k7_i = wna * h2_i * h5_i
        k8_i = wna * h11_i * h8_i
        x1_i = (k2_i + k3_i) * k5_i * k7_i + (k6_i + k7_i) * k2_i * k4_i
        x2_i = (k1_i + k8_i) * k4_i * k6_i + (k4_i + k5_i) * k1_i * k7_i
        x3_i = (k2_i + k3_i) * k6_i * k8_i + (k6_i + k7_i) * k1_i * k3_i
        x4_i = (k1_i + k8_i) * k3_i * k5_i + (k4_i + k5_i) * k2_i * k8_i
        E1_i = x1_i / (x1_i + x2_i + x3_i + x4_i)
        E2_i = x2_i / (x1_i + x2_i + x3_i + x4_i)
        E3_i = x3_i / (x1_i + x2_i + x3_i + x4_i)
        E4_i = x4_i / (x1_i + x2_i + x3_i + x4_i)
        allo_i = 1.0 / (1.0 + ufl.elem_pow(KmCaAct / cai, 2.0))
        zna = 1.0
        JncxNa_i = E3_i * k4pp_i - E2_i * k3pp_i + 3.0 * E4_i * k7_i - 3.0 * E1_i * k8_i
        JncxCa_i = E2_i * k2_i - E1_i * k1_i
        INaCa_i = (
            0.8
            * Gncx
            * scale_popu_KNCX
            * HF_scaling_Gncx
            * (zca * JncxCa_i + zna * JncxNa_i)
            * allo_i
        )

        # Expressions for the INaCa_ss component
        h1 = 1.0 + (1.0 + hna) * nass / kna3
        h2 = hna * nass / (kna3 * h1)
        h3 = 1.0 / h1
        h4 = 1.0 + (1 + nass / kna2) * nass / kna1
        h5 = (nass * nass) / (kna1 * kna2 * h4)
        h6 = 1.0 / h4
        h7 = 1.0 + nao * (1.0 + 1.0 / hna) / kna3
        h8 = nao / (kna3 * h7 * hna)
        h9 = 1.0 / h7
        h10 = 1.0 + kasymm + nao * (1 + nao / kna2) / kna1
        h11 = (nao * nao) / (kna1 * kna2 * h10)
        h12 = 1.0 / h10
        k1 = cao * kcaon * h12
        k2 = kcaoff
        k3p_ss = wca * h9
        k3pp = wnaca * h8
        k3 = k3p_ss + k3pp
        k4p_ss = wca * h3 / hca
        k4pp = wnaca * h2
        k4 = k4p_ss + k4pp
        k5 = kcaoff
        k6 = kcaon * cass * h6
        k7 = wna * h2 * h5
        k8 = wna * h11 * h8
        x1_ss = (k2 + k3) * k5 * k7 + (k6 + k7) * k2 * k4
        x2_ss = (k1 + k8) * k4 * k6 + (k4 + k5) * k1 * k7
        x3_ss = (k2 + k3) * k6 * k8 + (k6 + k7) * k1 * k3
        x4_ss = (k1 + k8) * k3 * k5 + (k4 + k5) * k2 * k8
        E1_ss = x1_ss / (x1_ss + x2_ss + x3_ss + x4_ss)
        E2_ss = x2_ss / (x1_ss + x2_ss + x3_ss + x4_ss)
        E3_ss = x3_ss / (x1_ss + x2_ss + x3_ss + x4_ss)
        E4_ss = x4_ss / (x1_ss + x2_ss + x3_ss + x4_ss)
        allo_ss = 1.0 / (1.0 + ufl.elem_pow(KmCaAct / cass, 2.0))
        JncxNa_ss = E3_ss * k4pp - E2_ss * k3pp + 3.0 * E4_ss * k7 - 3.0 * E1_ss * k8
        JncxCa_ss = E2_ss * k2 - E1_ss * k1
        INaCa_ss = (
            0.2
            * Gncx
            * scale_popu_KNCX
            * HF_scaling_Gncx
            * (zca * JncxCa_ss + zna * JncxNa_ss)
            * allo_ss
        )

        # Expressions for the INaK component
        Knai = Knai0 * ufl.exp(0.3333333333333333 * F * delta * v / (R * T))
        Knao = Knao0 * ufl.exp(0.3333333333333333 * F * (1.0 - delta) * v / (R * T))
        P = eP / (1.0 + H / Khp + nai / Knap + ki / Kxkur)
        a1 = (
            k1p
            * ufl.elem_pow(nai / Knai, 3.0)
            / (
                -1.0
                + ufl.elem_pow(1.0 + ki / Kki, 2.0)
                + ufl.elem_pow(1.0 + nai / Knai, 3.0)
            )
        )
        b1 = MgADP * k1m
        a2 = k2p
        b2 = (
            k2m
            * ufl.elem_pow(nao / Knao, 3.0)
            / (
                -1.0
                + ufl.elem_pow(1.0 + ko / Kko, 2.0)
                + ufl.elem_pow(1.0 + nao / Knao, 3.0)
            )
        )
        a3 = (
            k3p
            * ufl.elem_pow(ko / Kko, 2.0)
            / (
                -1.0
                + ufl.elem_pow(1.0 + ko / Kko, 2.0)
                + ufl.elem_pow(1.0 + nao / Knao, 3.0)
            )
        )
        b3 = H * k3m * P / (1.0 + MgATP / Kmgatp)
        a4 = MgATP * k4p / (Kmgatp * (1.0 + MgATP / Kmgatp))
        b4 = (
            k4m
            * ufl.elem_pow(ki / Kki, 2.0)
            / (
                -1.0
                + ufl.elem_pow(1.0 + ki / Kki, 2.0)
                + ufl.elem_pow(1.0 + nai / Knai, 3.0)
            )
        )
        x1 = a1 * a2 * a4 + a1 * a2 * b3 + a2 * b3 * b4 + b2 * b3 * b4
        x2 = a1 * a2 * a3 + a2 * a3 * b4 + a3 * b1 * b4 + b1 * b2 * b4
        x3 = a2 * a3 * a4 + a3 * a4 * b1 + a4 * b1 * b2 + b1 * b2 * b3
        x4 = a1 * a3 * a4 + a1 * a4 * b2 + a1 * b2 * b3 + b2 * b3 * b4
        E1 = x1 / (x1 + x2 + x3 + x4)
        E2 = x2 / (x1 + x2 + x3 + x4)
        E3 = x3 / (x1 + x2 + x3 + x4)
        E4 = x4 / (x1 + x2 + x3 + x4)
        JnakNa = 3.0 * E1 * a3 - 3.0 * E2 * b3
        JnakK = 2.0 * E4 * b1 - 2.0 * E3 * a1
        INaK = Pnak * scale_popu_KNaK * HF_scaling_Pnak * (zk * JnakK + zna * JnakNa)

        # Expressions for the IKb component
        xkb = 1.0 / (1.0 + 2.202363450949239 * ufl.exp(-0.05452562704471101 * v))
        IKb = GKb * scale_popu_GbK * scale_drug_IKb * (-EK + v) * xkb

        # Expressions for the INab component
        INab = (
            PNab
            * scale_drug_INab
            * scale_popu_GbNa
            * (-nao + ufl.exp(vfrt) * nai)
            * vffrt
            / (-1.0 + ufl.exp(vfrt))
        )

        # Expressions for the ICab component
        ICab = (
            4.0
            * PCab
            * scale_drug_ICab
            * scale_popu_GbCa
            * (-0.341 * cao + cai * ufl.exp(2.0 * vfrt))
            * vffrt
            / (-1.0 + ufl.exp(2.0 * vfrt))
        )

        # Expressions for the IpCa component
        IpCa = GpCa * scale_drug_IpCa * scale_popu_KpCa * cai / (0.0005 + cai)

        # Expressions for the Isac (Pueyo)--> ns + k component
        Gsac_ns = 0.006 * scale_drug_Isacns
        Esac_ns = -10.0
        lambda_max = 1.1
        Gsac_k = (0.2882 * 800.0 / 210.0) * scale_drug_Isack  # Pueyo endo
        Isac_P_ns = ufl.conditional(
            ufl.lt(lmbda, 1.0),
            0.0,
            Gsac_ns * ((lmbda - 1.0) / (lambda_max - 1.0)) * (v - Esac_ns),
        )
        Isac_P_k = ufl.conditional(
            ufl.lt(lmbda, 1.0),
            0.0,
            Gsac_k
            * ((lmbda - 1.0) / (lambda_max - 1.0))
            * (1.0 / (1.0 + ufl.exp((19.05 - v) / (29.98)))),
        )

        # Expressions for the Istim component
        Istim = 0  # amp*(ufl.le(time, duration))

        # Expressions for the diffusion fluxes component
        JdiffNa = 0.5 * nass - 0.5 * nai
        JdiffK = 0.5 * kss - 0.5 * ki
        Jdiff = 5.0 * cass - 5.0 * cai

        # Expressions for the ryanodine receptor component
        a_rel = 0.5 * bt
        Jrel_inf = (
            -ICaL
            * a_rel
            / (1.0 + 25.62890625 * HF_scaling_Jrel_inf * ufl.elem_pow(1.0 / cajsr, 8.0))
        )
        tau_rel_tmp = bt / (1.0 + 0.0123 / cajsr)
        tau_rel = ufl.conditional(ufl.lt(tau_rel_tmp, 0.001), 0.001, tau_rel_tmp)
        F_expressions[30] = (-Jrelnp + Jrel_inf) / tau_rel
        btp = 1.25 * bt
        a_relp = 0.5 * btp
        Jrel_infp = (
            -ICaL
            * a_relp
            / (1.0 + 25.62890625 * HF_scaling_Jrel_inf * ufl.elem_pow(1.0 / cajsr, 8.0))
        )
        tau_relp_tmp = btp / (1.0 + 0.0123 / cajsr)
        tau_relp = ufl.conditional(ufl.lt(tau_relp_tmp, 0.001), 0.001, tau_relp_tmp)
        F_expressions[31] = (-Jrelp + Jrel_infp) / tau_relp
        fJrelp = 1.0 / (1.0 + KmCaMK / CaMKa)
        Jrel = ((1.0 - fJrelp) * Jrelnp + Jrelp * fJrelp) * scale_popu_KRyR

        # Expressions for the calcium buffers component
        Jupnp = 0.004375 * cai / (0.00092 + cai)
        Jupp = 0.01203125 * cai / (0.00075 + cai)
        fJupp = 1.0 / (1.0 + KmCaMK / CaMKa)
        Jleak = 0.0002625 * cansr * scale_popu_Kleak * HF_scaling_Jleak
        Jup = (
            -Jleak
            + ((1.0 - fJupp) * Jupnp + Jupp * fJupp)
            * scale_popu_KSERCA
            * HF_scaling_Jup
        )
        Jtr = 0.01 * cansr - 0.01 * cajsr

        # Expressions for the intracellular concentrations component
        F_expressions[32] = JdiffNa * vss / vmyo + (
            -INa - INaL - INab - Isac_P_ns / 3.0 - 3.0 * INaCa_i - 3.0 * INaK
        ) * Acap / (F * vmyo)
        F_expressions[33] = -JdiffNa + (-ICaNa - 3.0 * INaCa_ss) * Acap / (F * vss)
        F_expressions[34] = JdiffK * vss / vmyo + (
            -Isac_P_k
            - IK1
            - IKb
            - IKr
            - IKs
            - Istim
            - Ito
            - Isac_P_ns / 3.0
            + 2.0 * INaK
        ) * Acap / (F * vmyo)
        F_expressions[35] = -JdiffK - Acap * ICaK / (F * vss)
        Bcass = 1.0 / (
            1.0
            + BSLmax * KmBSL * ufl.elem_pow(KmBSL + cass, -2.0)
            + BSRmax * KmBSR * ufl.elem_pow(KmBSR + cass, -2.0)
        )
        F_expressions[36] = (
            -Jdiff
            + Jrel * vjsr / vss
            + 0.5 * (-ICaL + 2.0 * INaCa_ss) * Acap / (F * vss)
        ) * Bcass
        F_expressions[37] = -Jtr * vjsr / vnsr + Jup
        Bcajsr = 1.0 / (1.0 + csqnmax * kmcsqn * ufl.elem_pow(kmcsqn + cajsr, -2.0))
        F_expressions[38] = (-Jrel + Jtr) * Bcajsr

        # Expressions for the mechanics component
        kwu = -kws * scale_popu_kws + (kuw * scale_popu_kuw) * (
            -1.0 + 1.0 / (rw * scale_popu_rw)
        )
        ksu = (
            kws
            * scale_popu_kws
            * rw
            * scale_popu_rw
            * (-1.0 + 1.0 / (rs * scale_popu_rs))
        )

        lambda_min12 = ufl.conditional(ufl.lt(lmbda, 1.2), lmbda, 1.2)
        XS = ufl.conditional(ufl.lt(XS, 0.0), 0.0, XS)
        XW = ufl.conditional(ufl.lt(XW, 0.0), 0.0, XW)

        XU = 1.0 - TmB - XS - XW
        gammawu = gammaw * abs(Zetaw)
        # gammasu = gammas*ufl.conditional(ufl.gt(Zetas*(ufl.gt(Zetas, 0)), (-1 -\
        #     Zetas)*(ufl.lt(Zetas, -1))), Zetas*(ufl.gt(Zetas, 0)), (-1 -\
        #     Zetas)*(ufl.lt(Zetas, -1)))
        zetas1 = Zetas * ufl.conditional(ufl.gt(Zetas, 0.0), 1.0, 0.0)
        zetas2 = (-1.0 - Zetas) * ufl.conditional(ufl.lt(Zetas, -1.0), 1.0, 0.0)
        gammasu = gammas * Max(zetas1, zetas2)

        F_expressions[39] = kws * scale_popu_kws * XW - XS * gammasu - XS * ksu
        F_expressions[40] = (
            kuw * scale_popu_kuw * XU
            - kws * scale_popu_kws * XW
            - XW * gammawu
            - XW * kwu
        )
        cat50 = (
            cat50_ref * scale_popu_CaT50ref + Beta1 * (-1.0 + lambda_min12)
        ) * HF_scaling_cat50_ref
        CaTrpn = ufl.conditional(ufl.lt(CaTrpn, 0.0), 0.0, CaTrpn)
        F_expressions[41] = (
            ktrpn
            * scale_popu_kTRPN
            * (
                -CaTrpn
                + ufl.elem_pow(1000.0 * cai / cat50, ntrpn * scale_popu_nTRPN)
                * (1.0 - CaTrpn)
            )
        )
        kb = (
            ku
            * scale_popu_ku
            * ufl.elem_pow(Trpn50 * scale_popu_TRPN50, (ntm * scale_popu_nTm))
            / (
                1.0
                - (rs * scale_popu_rs)
                - rw * scale_popu_rw * (1.0 - (rs * scale_popu_rs))
            )
        )

        F_expressions[42] = (
            ufl.conditional(
                ufl.lt(ufl.elem_pow(CaTrpn, -(ntm * scale_popu_nTm) / 2.0), 100.0),
                ufl.elem_pow(CaTrpn, -(ntm * scale_popu_nTm) / 2.0),
                100.0,
            )
            * XU
            * kb
            - ku
            * scale_popu_ku
            * ufl.elem_pow(CaTrpn, (ntm * scale_popu_nTm) / 2.0)
            * TmB
        )

        C = -1.0 + lambda_min12
        dCd = -Cd + C
        eta = ufl.conditional(ufl.lt(dCd, 0.0), etas, etal)
        F_expressions[43] = p_k * (-Cd + C) / eta
        Bcai = 1.0 / (1.0 + cmdnmax * kmcmdn * ufl.elem_pow(kmcmdn + cai, -2.0))
        J_TRPN = trpnmax * F_expressions[41]
        F_expressions[44] = (
            -J_TRPN
            + Jdiff * vss / vmyo
            - Jup * vnsr / vmyo
            + 0.5 * (-ICab - IpCa - Isac_P_ns / 3.0 + 2.0 * INaCa_i) * Acap / (F * vmyo)
        ) * Bcai

        # Return results
        return as_vector(F_expressions)

    def num_states(self):
        return 48

    def __str__(self):
        return "ORdmm_Land_em_coupling cardiac cell model"



import math
import typing
from pathlib import Path
from typing import NamedTuple

import cbcbeat
import dolfin
import pulse
from tqdm import tqdm

from . import config
from . import em_model
from . import ep_model
from . import geometry
from . import land_model
from . import mechanics_model
from . import save_load_functions as io
from . import utils
from .em_model import EMCoupling
from .ORdmm_Land import ORdmm_Land as CellModel

logger = utils.getLogger(__name__)


class EMState(NamedTuple):
    coupling: EMCoupling
    solver: cbcbeat.SplittingSolver
    mech_heart: pulse.MechanicsProblem
    geometry: geometry.BaseGeometry
    t0: float


def setup_EM_model(config: config.Config):

    geo = geometry.load_geometry(
        mesh_path=config.geometry_path,
        schema_path=config.geometry_schema_path,
    )

    coupling = em_model.EMCoupling(geo)

    # Set-up solver and time it
    solver = setup_ep_solver(
        dt=config.dt,
        coupling=coupling,
        cell_init_file=config.cell_init_file,
        drug_factors_file=config.drug_factors_file,
        popu_factors_file=config.popu_factors_file,
        disease_state=config.disease_state,
        PCL=config.PCL,
    )

    coupling.register_ep_model(solver)

    mech_heart = setup_mechanics_solver(
        coupling=coupling,
        geo=geo,
        bnd_rigid=config.bnd_rigid,
        cell_params=solver.ode_solver._model.parameters(),
        pre_stretch=config.pre_stretch,
        traction=config.traction,
        spring=config.spring,
        fix_right_plane=config.fix_right_plane,
        set_material=config.set_material,
        mechanics_ode_scheme=config.mechanics_ode_scheme,
        use_custom_newton_solver=config.mechanics_use_custom_newton_solver,
    )

    return EMState(
        coupling=coupling,
        solver=solver,
        mech_heart=mech_heart,
        geometry=geo,
        t0=0,
    )


def setup_mechanics_solver(
    coupling: em_model.EMCoupling,
    geo: geometry.BaseGeometry,
    cell_params,
    bnd_rigid: bool = config.Config.bnd_rigid,
    pre_stretch: typing.Optional[typing.Union[dolfin.Constant, float]] = None,
    traction: typing.Union[dolfin.Constant, float] = None,
    spring: typing.Union[dolfin.Constant, float] = None,
    fix_right_plane: bool = config.Config.fix_right_plane,
    mechanics_ode_scheme: land_model.Scheme = config.Config.mechanics_ode_scheme,
    set_material: str = "",
    linear_solver="mumps",
    use_custom_newton_solver: bool = config.Config.mechanics_use_custom_newton_solver,
    Zetas_prev=None,
    Zetaw_prev=None,
    lmbda_prev=None,
    state_prev=None,
):
    """Setup mechanics model with dirichlet boundary conditions or rigid motion."""
    logger.info("Set up mechanics model")

    # Use parameters from Biaxial test in Holzapfel 2019 (Table 1).
    material_parameters = dict(
        a=2.28,
        a_f=1.686,
        b=9.726,
        b_f=15.779,
        a_s=0.0,
        b_s=0.0,
        a_fs=0.0,
        b_fs=0.0,
    )

    active_model = land_model.LandModel(
        f0=geo.f0,
        s0=geo.s0,
        n0=geo.n0,
        eta=0,
        parameters=cell_params,
        XS=coupling.XS_mech,
        XW=coupling.XW_mech,
        mesh=coupling.mech_mesh,
        scheme=mechanics_ode_scheme,
        Zetas=Zetas_prev,
        Zetaw=Zetaw_prev,
        lmbda=lmbda_prev,
    )

    material = pulse.HolzapfelOgden(
        active_model=active_model,
        parameters=material_parameters,
    )

    if set_material == "Guccione":
        material_parameters = pulse.Guccione.default_parameters()
        material_parameters["CC"] = 2.0
        material_parameters["bf"] = 8.0
        material_parameters["bfs"] = 4.0
        material_parameters["bt"] = 2.0

        material = pulse.Guccione(
            params=material_parameters,
            active_model=active_model,
        )

    problem = mechanics_model.create_slab_problem(
        material=material,
        geo=geo,
        bnd_rigid=bnd_rigid,
        pre_stretch=pre_stretch,
        traction=traction,
        spring=spring,
        fix_right_plane=fix_right_plane,
        linear_solver=linear_solver,
        use_custom_newton_solver=use_custom_newton_solver,
    )

    if state_prev is not None:
        problem.state.assign(state_prev)

    problem.solve()
    coupling.register_mech_model(problem)

    total_dofs = problem.state.function_space().dim()
    logger.info("Mechanics model")
    utils.print_mesh_info(coupling.mech_mesh, total_dofs)

    return problem


def setup_ep_solver(
    dt,
    coupling,
    scheme=config.Config.ep_ode_scheme,
    theta=config.Config.ep_theta,
    preconditioner=config.Config.ep_preconditioner,
    cell_params=None,
    cell_inits=None,
    cell_init_file=None,
    drug_factors_file=None,
    popu_factors_file=None,
    disease_state=config.Config.disease_state,
    PCL=config.Config.PCL,
):
    ps = ep_model.setup_splitting_solver_parameters(
        theta=theta,
        preconditioner=preconditioner,
        dt=dt,
        scheme=scheme,
    )

    cell_params = ep_model.handle_cell_params(
        cell_params=cell_params,
        disease_state=disease_state,
        drug_factors_file=drug_factors_file,
        popu_factors_file=popu_factors_file,
    )

    cell_inits = ep_model.handle_cell_inits(
        cell_inits=cell_inits,
        cell_init_file=cell_init_file,
    )

    cell_inits["lmbda"] = coupling.lmbda_ep
    cell_inits["Zetas"] = coupling.Zetas_ep
    cell_inits["Zetaw"] = coupling.Zetaw_ep

    cellmodel = CellModel(init_conditions=cell_inits, params=cell_params)

    # Set-up cardiac model
    ep_heart = ep_model.setup_ep_model(cellmodel, coupling.ep_mesh, PCL=PCL)
    timer = dolfin.Timer("SplittingSolver: setup")

    solver = cbcbeat.SplittingSolver(ep_heart, params=ps)

    timer.stop()
    # Extract the solution fields and set the initial conditions
    (vs_, vs, vur) = solver.solution_fields()
    vs_.assign(cellmodel.initial_conditions())

    # Output some degrees of freedom
    total_dofs = vs.function_space().dim()
    logger.info("EP model")
    utils.print_mesh_info(coupling.ep_mesh, total_dofs)
    return solver


class Runner:
    def __init__(
        self,
        conf: typing.Optional[config.Config] = None,
        empty: bool = False,
        **kwargs,
    ) -> None:

        if conf is None:
            conf = config.Config()

        self._config = conf

        if empty:
            return

        self._state_path = Path(self._config.outdir).joinpath("state.h5")
        reset = not self._config.load_state
        if not reset and self._state_path.is_file():
            # Load state
            logger.info("Load previously saved state")
            coupling, ep_solver, mech_heart, geo, t0 = io.load_state(
                self._state_path,
                self._config.drug_factors_file,
                self._config.popu_factors_file,
                self._config.disease_state,
                self._config.PCL,  # Set bcl from cli
            )
        else:
            logger.info("Create a new state")
            # Create a new state
            coupling, ep_solver, mech_heart, geo, t0 = setup_EM_model(self._config)
        self.coupling: em_model.EMCoupling = coupling
        self.ep_solver: cbcbeat.SplittingSolver = ep_solver
        self.mech_heart: mechanics_model.MechanicsProblem = mech_heart
        self.geometry = geo
        self._t0: float = t0

        self._reset = reset

        self._setup_assigners()
        self.outdir = self._config.outdir

        logger.info(f"Starting at t0={self._t0}")

    @property
    def _dt(self):
        return self._config.dt

    @_dt.setter
    def _dt(self, value):
        self._config.dt = value

    @property
    def outdir(self):
        return self._outdir

    @outdir.setter
    def outdir(self, outdir):
        self._outdir = outdir
        self._state_path = Path(outdir).joinpath("state.h5")
        self._setup_datacollector()

    @property
    def t(self) -> float:
        if self._time_stepper is None:
            raise RuntimeError("Please create a time stepper before solving")
        return self._time_stepper.t

    @property
    def t0(self) -> float:
        return self._t0

    def create_time_stepper(
        self,
        T: float,
        use_ns: bool = True,
        st_progress: typing.Any = None,
    ) -> None:
        self._time_stepper = TimeStepper(
            t0=self._t0,
            T=T,
            dt=self._dt,
            use_ns=True,
            st_progress=st_progress,
        )

    @classmethod
    def from_models(
        cls,
        coupling: em_model.EMCoupling,
        ep_solver: cbcbeat.SplittingSolver,
        mech_heart: mechanics_model.MechanicsProblem,
        geo: geometry.BaseGeometry,
        reset: bool = True,
        t0: float = 0,
    ):
        obj = cls(empty=True)
        obj.coupling = coupling
        obj.ep_solver = ep_solver
        obj.mech_heart = mech_heart
        obj.geometry = geo
        obj._t0 = t0

        obj._reset = reset
        obj._dt = ep_solver.parameters["MonodomainSolver"]["default_timestep"]
        obj._setup_assigners()
        return obj

    def _setup_assigners(self):
        self._time_stepper = None
        self._vs = self.ep_solver.solution_fields()[1]
        self._v, self._v_assigner = utils.setup_assigner(self._vs, 0)
        self._Ca, self._Ca_assigner = utils.setup_assigner(self._vs, 45)
        self._XS, self._XS_assigner = utils.setup_assigner(self._vs, 40)
        self._XW, self._XW_assigner = utils.setup_assigner(self._vs, 41)
        self._CaTrpn, self._CaTrpn_assigner = utils.setup_assigner(self._vs, 42)
        self._TmB, self._TmB_assigner = utils.setup_assigner(self._vs, 43)
        self._Cd, self._Cd_assigner = utils.setup_assigner(self._vs, 44)
        self._Zetas, self._Zetas_assigner = utils.setup_assigner(self._vs, 47)
        self._Zetaw, self._Zetaw_assigner = utils.setup_assigner(self._vs, 48)

        self._pre_XS, self._preXS_assigner = utils.setup_assigner(self._vs, 40)
        self._pre_XW, self._preXW_assigner = utils.setup_assigner(self._vs, 41)

        self._u_subspace_index = self.mech_heart.u_subspace_index
        self._u, self._u_assigner = utils.setup_assigner(
            self.mech_heart.state,
            self._u_subspace_index,
        )
        self._assign_displacement()

    def _assign_displacement(self):
        self._u_assigner.assign(
            self._u,
            self.mech_heart.state.sub(self._u_subspace_index),
        )

    def _assign_ep(self):
        self._v_assigner.assign(self._v, utils.sub_function(self._vs, 0))
        self._Ca_assigner.assign(self._Ca, utils.sub_function(self._vs, 45))
        self._XS_assigner.assign(self._XS, utils.sub_function(self._vs, 40))
        self._XW_assigner.assign(self._XW, utils.sub_function(self._vs, 41))
        self._CaTrpn_assigner.assign(self._CaTrpn, utils.sub_function(self._vs, 42))
        self._TmB_assigner.assign(self._TmB, utils.sub_function(self._vs, 43))
        self._Cd_assigner.assign(self._Cd, utils.sub_function(self._vs, 44))
        self._Zetas_assigner.assign(self._Zetas, utils.sub_function(self._vs, 47))
        self._Zetaw_assigner.assign(self._Zetaw, utils.sub_function(self._vs, 48))

    def store(self):
        # Assign u, v and Ca for postprocessing
        self._assign_displacement()
        self._assign_ep()
        self.collector.store(TimeStepper.ns2ms(self.t))

    def _setup_datacollector(self):
        from .datacollector import DataCollector

        self.collector = DataCollector(
            outdir=self._outdir,
            geo=self.geometry,
            reset_state=self._reset,
        )
        for group, name, f in [
            ("mechanics", "u", self._u),
            ("ep", "V", self._v),
            ("ep", "Ca", self._Ca),
            ("mechanics", "lmbda", self.coupling.lmbda_mech),
            ("mechanics", "Ta", self.mech_heart.material.active.Ta_current),
            ("ep", "XS", self._XS),
            ("ep", "XW", self._XW),
            ("ep", "CaTrpn", self._CaTrpn),
            ("ep", "TmB", self._TmB),
            ("ep", "Cd", self._Cd),
            ("ep", "Zetas", self._Zetas),
            ("ep", "Zetaw", self._Zetaw),
            ("mechanics", "Zetas_mech", self.coupling.Zetas_mech),
            ("mechanics", "Zetaw_mech", self.coupling.Zetaw_mech),
            ("mechanics", "XS_mech", self.coupling.XS_mech),
            ("mechanics", "XW_mech", self.coupling.XW_mech),
        ]:
            self.collector.register(group, name, f)

    @property
    def dt_mechanics(self):
        return TimeStepper.ns2ms(float(self.t - self.mech_heart.material.active.t))

    def _solve_mechanics_now(self) -> bool:

        # Update these states that are needed in the Mechanics solver
        self.coupling.ep_to_coupling()

        # XS_norm = utils.compute_norm(self.coupling.XS_ep, self._pre_XS)
        # XW_norm = utils.compute_norm(self.coupling.XW_ep, self._pre_XW)

        # dt for the mechanics model should not be larger than 1 ms
        # return (XS_norm + XW_norm >= 0.05) #or self.dt_mechanics > 0.1
        return True  # self._t <= 10.0 or max(self.coupling.XS_ep.vector()) >= 0.0005 or max(self.coupling.XW_ep.vector()) >= 0.002

    def _pre_mechanics_solve(self) -> None:
        self._preXS_assigner.assign(self._pre_XS, utils.sub_function(self._vs, 40))
        self._preXW_assigner.assign(self._pre_XW, utils.sub_function(self._vs, 41))

        self.coupling.coupling_to_mechanics()
        self.mech_heart.material.active.update_time(self.t)

    def _post_mechanics_solve(self) -> None:

        # Update previous active tension
        self.mech_heart.material.active.update_prev()
        self.coupling.mechanics_to_coupling()
        self.coupling.coupling_to_ep()

    def _solve_mechanics(self):
        self._pre_mechanics_solve()
        if self._config.mechanics_use_continuation:
            self.mech_heart.solve_for_control(self.coupling.XS_ep)
        else:
            self.mech_heart.solve()
        # converged = False

        # current_t = float(self.mech_heart.material.active.t)
        # target_t = self._t
        # dt = self.dt_mechanics
        # while not converged:
        #     t = current_t + dt
        #     self.mech_heart.material.active.update_time(t)
        #     try:
        #         self.mech_heart.solve()
        #     except pulse.mechanicsproblem.SolverDidNotConverge:
        #         logger.warning(f"Failed to solve mechanics problem with dt={dt}")
        #         dt /= 2
        #         logger.warning(f"Try with dt={dt}")
        #         if dt < 1e-6:
        #             logger.warning("dt is too small. Good bye")
        #             raise

        #     else:
        #         if abs(t - target_t) < 1e-12:
        #             # We have reached the target
        #             converged = True
        #         # Update dt so that we hit the target next time
        #         current_t = t
        #         dt = target_t - t
        #         self.mech_heart.material.active.update_prev()

        self._post_mechanics_solve()

    def save_state(self, path):
        io.save_state(
            path,
            solver=self.ep_solver,
            mech_heart=self.mech_heart,
            geo=self.geometry,
            dt=self._dt,
            t0=TimeStepper.ns2ms(self.t),
        )

    def solve(
        self,
        T: float = config.Config.T,
        save_freq: int = config.Config.save_freq,
        show_progress_bar: bool = config.Config.show_progress_bar,
        st_progress: typing.Any = None,
    ):
        if not hasattr(self, "_outdir"):
            raise RuntimeError("Please set the output directory")

        # Truncate
        # if Path("residual.txt").is_file():
        #     fr = open("residual.txt", "w")
        #     fr.truncate(0)
        #     fr.close()

        save_it = int(save_freq / self._dt)
        self.create_time_stepper(T, use_ns=True, st_progress=st_progress)
        pbar = create_progressbar(
            time_stepper=self._time_stepper,
            show_progress_bar=show_progress_bar,
        )

        # Store initial state
        # save_state_every_n_beat = 5  # Save state every fifth beat
        # five_beats = (
        #     TimeStepper.ms2ns(save_state_every_n_beat * 1000.0) / self._time_stepper.dt
        # )
        # beat_nr = 0
        self.mech_heart.material.active.start_time(self.t)

        for (i, (t0, t)) in enumerate(pbar):
            logger.debug(
                f"Solve EP model at step {i} from {TimeStepper.ns2ms(t0):.2f} ms to {TimeStepper.ns2ms(t):.2f} ms",
            )

            # Solve EP model
            self.ep_solver.step((TimeStepper.ns2ms(t0), TimeStepper.ns2ms(t)))

            if self._solve_mechanics_now():
                logger.debug(
                    (
                        f"Solve mechanics model at step {i} from "
                        f"{TimeStepper.ns2ms(self.mech_heart.material.active.t):.2f} ms"
                        f" to {TimeStepper.ns2ms(self.t):.2f} ms with timestep "
                        f"{self.dt_mechanics:.5f} ms"
                    ),
                )
                self._solve_mechanics()

            self.ep_solver.vs_.assign(self.ep_solver.vs)

            # Store every 'save_freq' ms
            if i % save_it == 0:
                self.store()

            # Store state every 5 beats
            # if i > 0 and (i + 1) % five_beats == 0:
            #     self.save_state(
            #         self._state_path.parent.joinpath(
            #             f"state_{beat_nr}beat.h5",
            #         ),
            #     )
            #     beat_nr += save_state_every_n_beat

            # # Residual file : End of line after each time step
            # if (
            #     Path("residual.txt").is_file()
            #     and dolfin.MPI.rank(dolfin.MPI.comm_world) == 0
            # ):
            #     fr = open("residual.txt", "a")
            #     fr.write("\n")
            #     fr.close()

        self.save_state(self._state_path)

        # Copy residual file to output dir (if exists)
        # if Path("residual.txt").is_file():
        #     Path(self._outdir).joinpath("residual.txt").write_text(
        #         Path("residual.txt").read_text(),
        #     )


class _tqdm:
    def __init__(self, iterable, *args, **kwargs):
        self._iterable = iterable

    def set_postfix(self, msg):
        pass

    def __iter__(self):
        return iter(self._iterable)


class TimeStepper:
    def __init__(
        self,
        *,
        t0: float,
        T: float,
        dt: float,
        use_ns: bool = True,
        st_progress: typing.Any = None,
    ) -> None:
        """Initialize time stepper

        Parameters
        ----------
        t0 : float
            Start time in milliseconds
        T : float
            End time in milliseconds
        dt : float
            Time step
        use_ns : bool, optional
            Whether to return the time in nanoseconds, by default True
        st_progress:
            Streamlit progress bar
        """

        self._use_ns = use_ns
        self._st_progress = st_progress

        if use_ns:
            self.t0 = TimeStepper.ms2ns(t0)
            self.T = TimeStepper.ms2ns(T)
            self.dt = TimeStepper.ms2ns(dt)
        else:
            self.t0 = t0
            self.T = T
            self.dt = dt

        self.reset()

    def reset(self):
        self.t = self.t0
        self.step = 0

    @property
    def T(self) -> float:
        return self._T

    @T.setter
    def T(self, T: float) -> None:
        if self.t0 > T:
            raise ValueError("Start time has to be lower than end time")
        self._T = T

    @property
    def dt(self) -> float:
        return self._dt

    @dt.setter
    def dt(self, dt: float) -> None:
        dt = min(self.T - self.t0, dt)
        self._dt = dt

    @property
    def total_steps(self) -> int:
        if math.isclose(self.T, self.t0):
            return 0
        return round((self.T - self.t0) / self.dt)

    def __iter__(self):
        if self.T is None:
            raise RuntimeError("Please assign an end time to time stepper")
        while self.t < self.T:

            prev_t = self.t
            self.t = min(self.t + self.dt, self.T)
            self.step += 1
            if self._st_progress is not None:
                self._st_progress.progress(self.step / self.total_steps)
            yield prev_t, self.t

    @staticmethod
    def ns2ms(t: float) -> float:
        """Convert nanoseconds to milliseconds

        Parameters
        ----------
        t : float
            The time in nanoseconds

        Returns
        -------
        float
            Time in milliseconds
        """
        return t * 1e-6

    @staticmethod
    def ms2ns(t: float) -> float:
        """Convert from milliseconds to nanoseconds

        Parameters
        ----------
        t : float
            Time in milliseconds

        Returns
        -------
        float
            Time in nanoseconds
        """
        return int(t * 1e6)


def create_progressbar(
    time_stepper: TimeStepper,
    show_progress_bar: bool = config.Config.show_progress_bar,
):
    if show_progress_bar:
        # Show progressbar
        pbar = tqdm(time_stepper, total=time_stepper.total_steps)
    else:
        # Hide progressbar
        pbar = _tqdm(time_stepper, total=time_stepper.total_steps)
    return pbar



import typing

import dolfin
import pulse

from . import config
from . import geometry
from . import utils


def create_slab_boundary_conditions(
    geo: geometry.SlabGeometry,
    pre_stretch: typing.Optional[typing.Union[dolfin.Constant, float]] = None,
    traction: typing.Union[dolfin.Constant, float] = None,
    spring: typing.Union[dolfin.Constant, float] = None,
    fix_right_plane: bool = config.Config.fix_right_plane,
) -> pulse.BoundaryConditions:
    """Completely fix the left side of the mesh in the x-direction (i.e the side with the
    lowest x-values), fix the plane with y=0 in the y-direction, fix the plane with
    z=0 in the z-direction and apply some boundary condition to the right side.


    Parameters
    ----------
    geo : simcardems.geometry.SlabGeometry
        A slab geometry
    pre_stretch : typing.Union[dolfin.Constant, float], optional
        Value representing the amount of pre stretch, by default None
    traction : typing.Union[dolfin.Constant, float], optional
        Value representing the amount of traction, by default None
    spring : typing.Union[dolfin.Constant, float], optional
        Value representing the stiffness of the string, by default None
    fix_right_plane : bool, optional
        Fix the right plane so that it is not able to move in any direction
        except the x-direction, by default False


    Notes
    -----
    If `pre_stretch` if different from None, a pre stretch will be applied
    to the right side, through a Dirichlet boundary condition.

    If `traction` is different from None then an external force is applied
    to the right size, through a Neumann boundary condition.
    A positive value means that the force is compressing while a
    negative value means that it is stretching

    If `spring` is different from None then the amount of force that needs
    to be applied to displace the right boundary increases with the amount
    of displacement. The value of `spring` represents the stiffness of the
    spring.

    """

    def dirichlet_bc(W):
        # BC with fixing left size
        bcs = [
            dolfin.DirichletBC(
                W.sub(0).sub(0),
                dolfin.Constant(0.0),
                geo.ffun,
                geo.markers["X0"][0],
            ),
            dolfin.DirichletBC(
                W.sub(0).sub(1),  # u_y
                dolfin.Constant(0.0),
                geo.ffun,
                geo.markers["Y0"][0],
            ),
            dolfin.DirichletBC(
                W.sub(0).sub(2),  # u_z
                dolfin.Constant(0.0),
                geo.ffun,
                geo.markers["Z0"][0],
            ),
        ]
        if fix_right_plane:
            bcs.append(
                dolfin.DirichletBC(
                    W.sub(0).sub(0),  # u_x
                    dolfin.Constant(0.0),
                    geo.ffun,
                    geo.markers["X1"][0],
                ),
            )

        if pre_stretch is not None:
            bcs.append(
                dolfin.DirichletBC(
                    W.sub(0).sub(0),
                    utils.float_to_constant(pre_stretch),
                    geo.ffun,
                    geo.markers["X1"][0],
                ),
            )
        return bcs

    neumann_bc = []
    if traction is not None:
        neumann_bc.append(
            pulse.NeumannBC(
                traction=utils.float_to_constant(traction),
                marker=geo.markers["X1"][0],
            ),
        )

    robin_bc = []
    if spring is not None:
        robin_bc.append(
            pulse.RobinBC(
                value=utils.float_to_constant(spring),
                marker=geo.markers["X1"][0],
            ),
        )

    return pulse.BoundaryConditions(
        dirichlet=(dirichlet_bc,),
        neumann=neumann_bc,
        robin=robin_bc,
    )


def create_lv_boundary_conditions(
    geo: geometry.LeftVentricularGeometry,
    traction: typing.Union[dolfin.Constant, float] = None,
    spring: typing.Union[dolfin.Constant, float] = None,
):
    def dirichlet_bc(W):
        # Completely fix the base
        return [
            dolfin.DirichletBC(
                W.sub(0),
                dolfin.Constant((0.0, 0.0, 0.0)),
                geo.ffun,
                geo.markers["BASE"][0],
            ),
        ]

    neumann_bc = []
    if traction is not None:
        # LV pressure
        neumann_bc.append(
            pulse.NeumannBC(
                traction=utils.float_to_constant(traction),
                marker=geo.markers["ENDO"][0],
            ),
        )

    robin_bc = []
    if spring is not None:
        # Pericardium
        robin_bc.append(
            pulse.RobinBC(
                value=utils.float_to_constant(spring),
                marker=geo.markers["EPI"][0],
            ),
        )

    return pulse.BoundaryConditions(
        dirichlet=(dirichlet_bc,),
        neumann=neumann_bc,
        robin=robin_bc,
    )



import abc
import typing
from enum import Enum

import dolfin
import numpy as np

from .geometry import BaseGeometry
from .geometry import LeftVentricularGeometry
from .geometry import SlabGeometry


class ValueExtractor:
    def __init__(self, geo: BaseGeometry):
        self.geo = geo
        self.volume = dolfin.assemble(dolfin.Constant(1.0) * dolfin.dx(domain=geo.mesh))

        if isinstance(self.geo, SlabGeometry):
            self.boundary: Boundary = SlabBoundary(geo.mesh)
        elif isinstance(self.geo, LeftVentricularGeometry):
            self.boundary = LVBoundary(geo.mesh)
        else:
            raise NotImplementedError

    def average(self, func: dolfin.Function) -> float:
        if func.value_rank() == 0:
            return dolfin.assemble(func * dolfin.dx) / self.volume
        else:
            # Take the magnitude
            return dolfin.assemble(dolfin.sqrt(func**2) * dolfin.dx) / self.volume

    def eval_at_node(
        self,
        func: dolfin.Function,
        point: np.ndarray,
        dofs: typing.Optional[np.ndarray] = None,
    ):
        try:
            return func(point)
        except RuntimeError as e:
            if dofs is None:
                msg = "Unable to evaluate function at Node"
                raise RuntimeError(msg) from e
            closest_dof = np.argmin(
                np.linalg.norm(dofs - np.array(point), axis=1),
            )
            return func.vector().get_local()[closest_dof]

    def eval(
        self,
        func: dolfin.Function,
        value: str,
        dofs: typing.Optional[np.ndarray] = None,
    ):

        if value == "average":
            return self.average(func)
        elif value in self.boundary.nodes():
            point = getattr(self.boundary, value)
            return self.eval_at_node(func=func, point=point, dofs=dofs)

        msg = f"Value {value} is not implemented"
        raise NotImplementedError(msg)


def center_func(fmin, fmax):
    return fmin + (fmax - fmin) / 2


class Boundary(abc.ABC):
    def __init__(self, mesh):
        self.mesh = mesh

    @staticmethod
    @abc.abstractmethod
    def nodes() -> typing.Sequence[str]:
        ...


class LVBoundary(Boundary):
    @staticmethod
    def nodes():
        return []


class SlabBoundaryNodes(Enum):
    center = "center"
    xmax = "xmax"
    xmin = "xmin"
    ymax = "ymax"
    ymin = "ymin"
    zmax = "zmax"
    zmin = "zmin"


class SlabBoundary(Boundary):
    @staticmethod
    def nodes():
        return SlabBoundaryNodes._member_names_

    @property
    def boundaries(self):
        coords = self.mesh.coordinates()
        return dict(
            max_x=coords.T[0].max(),
            min_x=coords.T[0].min(),
            max_y=coords.T[1].max(),
            min_y=coords.T[1].min(),
            max_z=coords.T[2].max(),
            min_z=coords.T[2].min(),
        )

    @property
    def xmin(self):
        return [
            self.boundaries["min_x"],
            center_func(self.boundaries["min_y"], self.boundaries["max_y"]),
            center_func(self.boundaries["min_z"], self.boundaries["max_z"]),
        ]

    @property
    def xmax(self):
        return [
            self.boundaries["max_x"],
            center_func(self.boundaries["min_y"], self.boundaries["max_y"]),
            center_func(self.boundaries["min_z"], self.boundaries["max_z"]),
        ]

    @property
    def ymin(self):
        return [
            center_func(self.boundaries["min_x"], self.boundaries["max_x"]),
            self.boundaries["min_y"],
            center_func(self.boundaries["min_z"], self.boundaries["max_z"]),
        ]

    @property
    def ymax(self):
        return [
            center_func(self.boundaries["min_x"], self.boundaries["max_x"]),
            self.boundaries["max_y"],
            center_func(self.boundaries["min_z"], self.boundaries["max_z"]),
        ]

    @property
    def zmin(self):
        return [
            center_func(self.boundaries["min_x"], self.boundaries["max_x"]),
            center_func(self.boundaries["min_y"], self.boundaries["max_y"]),
            self.boundaries["min_z"],
        ]

    @property
    def zmax(self):
        return [
            center_func(self.boundaries["min_x"], self.boundaries["max_x"]),
            center_func(self.boundaries["min_y"], self.boundaries["max_y"]),
            self.boundaries["max_z"],
        ]

    @property
    def center(self):
        return [
            center_func(self.boundaries["min_x"], self.boundaries["max_x"]),
            center_func(self.boundaries["min_y"], self.boundaries["max_y"]),
            center_func(self.boundaries["min_z"], self.boundaries["max_z"]),
        ]



from enum import Enum
from pathlib import Path
from typing import Dict
from typing import List
from typing import Union

import dolfin
from dolfin import FiniteElement  # noqa: F401
from dolfin import tetrahedron  # noqa: F401
from dolfin import VectorElement  # noqa: F401

from . import utils
from .geometry import BaseGeometry
from .geometry import load_geometry
from .save_load_functions import h5pyfile
from .value_extractor import ValueExtractor

logger = utils.getLogger(__name__)


class DataGroups(Enum):
    ep = "ep"
    mechanics = "mechanics"


class DataCollector:
    def __init__(self, outdir, geo: BaseGeometry, reset_state=True) -> None:
        self.outdir = Path(outdir)
        self._results_file = self.outdir.joinpath("results.h5")
        self.comm = geo.mesh.mpi_comm()  # FIXME: Is this important?

        if reset_state:
            utils.remove_file(self._results_file)

        self._times_stamps = set()
        if not self._results_file.is_file():
            geo.dump(self.results_file)

        else:

            try:
                with h5pyfile(self._results_file, "r") as f:
                    self._times_stamps = set(f["ep"]["V"].keys())
            except KeyError:
                pass

        self._functions: Dict[str, Dict[str, dolfin.Function]] = {
            "ep": {},
            "mechanics": {},
        }

    @property
    def results_file(self):
        return self._results_file.as_posix()

    def register(self, group: str, name: str, f: dolfin.Function) -> None:
        assert group in [
            "ep",
            "mechanics",
        ], f"Group has to be 'ep' or 'mechanics', got {group}"

        if name in self.names[group]:
            logger.warning(
                f"Warning: {name} in group {group} is already registered - overwriting",
            )
        self._functions[group][name] = f

    @property
    def names(self) -> Dict[str, List[str]]:
        return {k: list(v.keys()) for k, v in self._functions.items()}

    def store(self, t: float) -> None:

        t_str = f"{t:.2f}"
        logger.debug(f"Store results at time {t_str}")
        if f"{t_str}" in self._times_stamps:
            logger.info(f"Time stamp {t_str} already exist in file")
            return

        self._times_stamps.add(t_str)

        with dolfin.HDF5File(self.comm, self.results_file, "a") as h5file:
            for group, names in self.names.items():
                logger.debug(f"Save HDF5File {self.results_file} for group {group}")
                for name in names:
                    logger.debug(f"Save {name}")
                    f = self._functions[group][name]
                    h5file.write(f, f"{group}/{name}/{t_str}")


class DataLoader:
    def __init__(self, h5name) -> None:

        self._h5file = None
        self._h5name = Path(h5name)
        if not self._h5name.is_file():
            raise FileNotFoundError(f"File {h5name} does not exist")

        self.geo = load_geometry(self._h5name)

        with h5pyfile(self._h5name) as h5file:

            # Find the remaining functions
            self.names = {
                group: [name for name in h5file.get(group, {}).keys()]
                for group in ["ep", "mechanics"]
            }
            if len(self.names["ep"]) + len(self.names["mechanics"]) == 0:
                raise ValueError("No functions found in results file")

            # Get time stamps
            all_time_stamps = {
                "{group}:{name}": sorted(
                    list(h5file[group][name].keys()),
                    key=lambda x: float(x),
                )
                for group, names in self.names.items()
                for name in names
            }

            # An verify that they are all the same
            self.time_stamps = all_time_stamps[next(iter(all_time_stamps.keys()))]
            for name in all_time_stamps.keys():
                assert self.time_stamps == all_time_stamps[name], name

            if self.time_stamps is None or len(self.time_stamps) == 0:
                raise ValueError("No time stamps found")

            # Get the signatures - FIXME: Add a check that the signature exist.
            self._signatures = {
                group: {
                    name: h5file[group][name][self.time_stamps[0]]
                    .attrs["signature"]
                    .decode()
                    for name in names
                }
                for group, names in self.names.items()
            }

        self._h5file = dolfin.HDF5File(
            self.ep_mesh.mpi_comm(),
            self._h5name.as_posix(),
            "r",
        )

        self._create_functions()
        self.value_extractor = ValueExtractor(self.geo)

    @property
    def ep_mesh(self):
        return self.geo.ep_mesh

    @property
    def mech_mesh(self):
        return self.geo.mechanics_mesh

    @property
    def size(self) -> int:
        return len(self.time_stamps)

    def __repr__(self):
        return f"{self.__class__.__name__}({str(self._h5name)})"

    def __del__(self):
        if self._h5file is not None:
            self._h5file.close()

    def _create_functions(self):
        self._function_spaces = {}

        for group, signature_dict in self._signatures.items():
            mesh = self.ep_mesh if group == "ep" else self.mech_mesh

            self._function_spaces.update(
                {
                    group: {
                        signature: dolfin.FunctionSpace(mesh, eval(signature))
                        for signature in set(signature_dict.values())
                    },
                },
            )

        self._functions = {
            group: {
                name: dolfin.Function(
                    self._function_spaces[group][self._signatures[group][name]],
                )
                for name in names
            }
            for group, names in self.names.items()
        }

        self._dofs = {
            group: {
                name: func.function_space().tabulate_dof_coordinates()
                for name, func in functions.items()
            }
            for group, functions in self._functions.items()
        }

    def extract_value(
        self,
        group: DataGroups,
        name: str,
        t: Union[str, float],
        reduction: str,
    ):
        func = self.get(group, name, t)
        dofs = self._dofs[self._group_to_str(group)][name]
        return self.value_extractor.eval(func, value=reduction, dofs=dofs)

    def _group_to_str(self, group):
        group_str = utils.enum2str(group, DataGroups)
        if group_str not in self.names:
            raise KeyError(
                f"Cannot find group {group} in names, expected of of {self.names.keys()}",
            )
        return group_str

    def get(
        self,
        group: DataGroups,
        name: str,
        t: Union[str, float],
    ) -> dolfin.Function:
        """Retrieve the function from the file

        Parameters
        ----------
        group : DataGroups
            The group where the function is stored, either
            'ep' or 'mechanics'
        name : str
            Name of the function
        t : Union[str, float]
            Time stamp you want to use. See `DataLoader.time_stamps`

        Returns
        -------
        dolfin.Function
            The function

        Raises
        ------
        KeyError
            If group does not exist in the file
        KeyError
            If name does not exist in group
        KeyError
            If 't' provided is not among the time stamps
        """
        if self._h5file is None:
            raise RuntimeError(f"Unable to read from file {self._h5name}")

        group_str = self._group_to_str(group)
        names = self.names[group_str]
        if f"{name}" not in names:
            raise KeyError(
                f"Cannot find name {name} in group {group}. Possible options are {names}",
            )

        if isinstance(t, (int, float)):
            t = f"{t:.2f}"
        if t not in self.time_stamps:
            raise KeyError(f"Invalid time stamps {t}")

        func = self._functions[group_str][name]
        self._h5file.read(func, f"{group_str}/{name}/{t}/")
        return func



import json
import typing
from pathlib import Path

import click
import dolfin

from . import config
from . import land_model
from . import postprocess as post
from . import utils
from .setup_models import Runner
from .version import __version__

logger = utils.getLogger(__name__)


@click.group()
@click.version_option(__version__)
def cli():
    pass


@click.command("run")
@click.argument("geometry-path", type=click.Path(exists=True, resolve_path=True))
@click.option(
    "-s",
    "--geometry-schema-path",
    default=None,
    help=(
        "Schema for the geometry. If not provided it will assume that this file has the "
        "same file name as `geometry-path` but with the suffix `.json`."
    ),
)
@click.option(
    "-o",
    "--outdir",
    default=config.Config.outdir,
    type=click.Path(writable=True, resolve_path=True),
    help="Output directory",
)
@click.option("--dt", default=config.Config.dt, type=float, help="Time step")
@click.option(
    "-T",
    "--end-time",
    "T",
    default=config.Config.T,
    type=float,
    help="End-time of simulation",
)
@click.option(
    "-n",
    "--num_refinements",
    default=config.Config.num_refinements,
    type=int,
    help="Number of refinements of for the mesh using in the EP model",
)
@click.option(
    "--save_freq",
    default=config.Config.save_freq,
    type=int,
    help="Set frequency of saving results to file",
)
@click.option(
    "--set_material",
    default=config.Config.set_material,
    type=str,
    help="Choose material properties for mechanics model (default is HolzapfelOgden, option is Guccione",
)
@click.option(
    "--bnd_rigid",
    is_flag=True,
    default=config.Config.bnd_rigid,
    help="Flag to set boundary conditions for the mechanics problem to rigid motion condition",
)
@click.option(
    "--load_state",
    is_flag=True,
    default=config.Config.load_state,
    help="If load existing state if exists, otherwise create a new state",
)
@click.option(
    "-IC",
    "--cell_init_file",
    default=config.Config.cell_init_file,
    type=click.Path(),
    help=(
        "Path to file containing initial conditions (json or h5 file). "
        "If none is provided then the default initial conditions will be used"
    ),
)
@click.option(
    "--loglevel",
    default=config.Config.loglevel,
    type=int,
    help="How much printing. DEBUG: 10, INFO:20 (default), WARNING: 30",
)
@click.option(
    "--show_progress_bar/--hide_progress_bar",
    default=config.Config.show_progress_bar,
    help="Shows or hide the progress bar.",
)
@click.option(
    "--drug_factors_file",
    default=config.Config.drug_factors_file,
    type=click.Path(),
    help="Path to drugs scaling factors file (json)",
)
@click.option(
    "--popu_factors_file",
    default=config.Config.popu_factors_file,
    type=click.Path(),
    help="Path to population scaling factors file (json)",
)
@click.option(
    "--disease_state",
    default=config.Config.disease_state,
    type=str,
    help="Indicate disease state. Default is healthy. ",
)
@click.option(
    "--mechanics-ode-scheme",
    default=config.Config.mechanics_ode_scheme,
    type=click.Choice(land_model.Scheme._member_names_),
    help="Scheme used to solve the ODEs in the mechanics model",
)
@click.option(
    "--mechanics-use-continuation",
    default=config.Config.mechanics_use_continuation,
    type=bool,
    help="Use continuation based mechanics solver",
)
@click.option(
    "--mechanics-use-custom-newton-solver",
    default=config.Config.mechanics_use_custom_newton_solver,
    type=bool,
    help="Use custom newton solver and solve ODEs at each Newton iteration",
)
@click.option(
    "--pcl",
    default=config.Config.PCL,
    type=float,
    help="Pacing cycle length (ms)",
)
@click.option(
    "--spring",
    default=config.Config.spring,
    type=float,
    help="Set value of spring for Robin boundary condition",
)
@click.option(
    "--traction",
    default=config.Config.traction,
    type=float,
    help="Set value of traction for Neumann boundary condition",
)
@click.option(
    "--fix_right_plane",
    is_flag=True,
    default=config.Config.fix_right_plane,
    help="Fix right plane in fiber direction (only usable for slab)",
)
def run(
    geometry_path: utils.PathLike,
    geometry_schema_path: typing.Optional[utils.PathLike],
    outdir: utils.PathLike,
    T: float,
    dt: float,
    bnd_rigid: bool,
    load_state: bool,
    cell_init_file: utils.PathLike,
    show_progress_bar: bool,
    save_freq: int,
    loglevel: int,
    num_refinements: int,
    set_material: str,
    drug_factors_file: utils.PathLike,
    popu_factors_file: utils.PathLike,
    disease_state: str,
    mechanics_ode_scheme: land_model.Scheme,
    mechanics_use_continuation: bool,
    mechanics_use_custom_newton_solver: bool,
    pcl: float,
    spring: float,
    traction: float,
    fix_right_plane: bool,
):
    conf = config.Config(
        geometry_path=geometry_path,
        geometry_schema_path=geometry_schema_path,
        outdir=outdir,
        T=T,
        dt=dt,
        bnd_rigid=bnd_rigid,
        spring=spring,
        traction=traction,
        fix_right_plane=fix_right_plane,
        load_state=load_state,
        cell_init_file=cell_init_file,
        show_progress_bar=show_progress_bar,
        save_freq=save_freq,
        loglevel=loglevel,
        num_refinements=num_refinements,
        set_material=set_material,
        drug_factors_file=drug_factors_file,
        popu_factors_file=popu_factors_file,
        disease_state=disease_state,
        mechanics_ode_scheme=mechanics_ode_scheme,
        mechanics_use_continuation=mechanics_use_continuation,
        mechanics_use_custom_newton_solver=mechanics_use_custom_newton_solver,
        PCL=pcl,
    )
    main(conf=conf)


@click.command("run-json")
@click.argument("path", required=True, type=click.Path(exists=True))
def run_json(path):
    with open(path, "r") as json_file:
        data = json.load(json_file)

    main(config.Config(**data))


def main(conf: typing.Optional[config.Config]):

    if conf is None:
        conf = config.Config()

    geometry_path = Path(conf.geometry_path)
    if not geometry_path.is_file():
        msg = f"Unable to to find geometry path {geometry_path}"
        raise IOError(msg)
    if conf.geometry_schema_path is None:
        conf.geometry_schema_path = geometry_path.with_suffix(".json")

    # Get all arguments and dump them to a json file
    info_dict = conf.__dict__
    outdir = Path(conf.outdir)
    outdir.mkdir(exist_ok=True)
    with open(outdir.joinpath("parameters.json"), "w") as f:
        json.dump(info_dict, f, default=post.json_serial)

    # Disable warnings
    from . import set_log_level

    set_log_level(conf.loglevel)
    dolfin.set_log_level(conf.loglevel + 10)  # TODO: Make it possible to set this?

    runner = Runner(conf=conf)

    runner.solve(
        T=conf.T,
        save_freq=conf.save_freq,
        show_progress_bar=conf.show_progress_bar,
    )


@click.command("postprocess")
@click.argument("folder", required=True, type=click.Path(exists=True))
@click.option(
    "--plot-state-traces",
    is_flag=True,
    default=False,
    help="Plot state traces",
)
@click.option(
    "--make-xdmf",
    is_flag=True,
    default=False,
    help="Make xdmf files",
)
@click.option(
    "--population",
    is_flag=True,
    default=False,
    help="Plot population",
)
@click.option(
    "--reduction",
    default="average",
    help="What type of reduction to perform when plotting state traces, by default 'average'.",
)
@click.option("--num_models", default=5, help="Number of models to be analyzed")
def postprocess(
    folder,
    num_models,
    plot_state_traces,
    make_xdmf,
    population,
    reduction,
):
    folder = Path(folder)
    if plot_state_traces:
        post.plot_state_traces(folder.joinpath("results.h5"), reduction=reduction)
    if make_xdmf:
        post.make_xdmffiles(folder.joinpath("results.h5"))
    if population:
        print("Execute postprocess for population")
        post.save_popu_json(folder, num_models)


@click.command("gui")
def gui():
    # Make sure we can import the required packages
    from . import gui  # noqa: F401

    gui_path = Path(__file__).parent.joinpath("gui.py")
    import subprocess as sp

    sp.run(["streamlit", "run", gui_path.as_posix()])


@click.command("run-benchmark")
@click.argument(
    "outdir",
    required=True,
    type=click.Path(resolve_path=True, writable=True),
)
@click.option(
    "--overwrite/--no-overwrite",
    default=True,
    type=bool,
    help="If True overwrite results in output directory",
)
def run_benchmark(outdir, overwrite):
    # Make sure we can import the required packages
    from . import benchmark  # noqa: F401

    benchmark_path = Path(__file__).parent.joinpath("benchmark.py")
    import subprocess as sp

    path = Path(outdir)
    if path.exists():
        if overwrite:
            import shutil

            shutil.rmtree(path)

    path.mkdir(parents=True, exist_ok=True)

    sp.run(["python3", benchmark_path, path.as_posix()])


cli.add_command(run)
cli.add_command(run_json)
cli.add_command(postprocess)
cli.add_command(gui)
cli.add_command(run_benchmark)



from enum import Enum

import dolfin
import pulse
import ufl

from .utils import Projector


class Scheme(str, Enum):
    fd = "fd"
    bd = "bd"
    analytic = "analytic"


def _Zeta(Zeta_prev, A, c, dLambda, dt, scheme: Scheme):

    if scheme == Scheme.analytic:
        return Zeta_prev * dolfin.exp(-c * dt) + (A * dLambda / c * dt) * (
            1.0 - dolfin.exp(-c * dt)
        )

    elif scheme == Scheme.bd:
        return Zeta_prev + A * dLambda / (1.0 + c * dt)
    else:
        return Zeta_prev * (1.0 - c * dt) + A * dLambda


class LandModel(pulse.ActiveModel):
    def __init__(
        self,
        XS,
        XW,
        parameters,
        mesh,
        Zetas=None,
        Zetaw=None,
        lmbda=None,
        f0=None,
        s0=None,
        n0=None,
        eta=0,
        scheme: Scheme = Scheme.analytic,
        **kwargs,
    ):
        super().__init__(f0=f0, s0=s0, n0=n0)
        self._eta = eta
        self.function_space = dolfin.FunctionSpace(mesh, "CG", 1)

        self.XS = XS
        self.XW = XW
        self._parameters = parameters
        self._t = 0.0
        self._t_prev = 0.0
        self._scheme = scheme

        self._dLambda = dolfin.Function(self.function_space)
        self.lmbda_prev = dolfin.Function(self.function_space)
        self.lmbda_prev.vector()[:] = 1.0
        if lmbda is not None:
            self.lmbda_prev.assign(lmbda)
        self.lmbda = dolfin.Function(self.function_space)

        self._Zetas = dolfin.Function(self.function_space)
        self.Zetas_prev = dolfin.Function(self.function_space)
        if Zetas is not None:
            self.Zetas_prev.assign(Zetas)

        self._Zetaw = dolfin.Function(self.function_space)
        self.Zetaw_prev = dolfin.Function(self.function_space)
        if Zetaw is not None:
            self.Zetaw_prev.assign(Zetaw)

        self.Ta_current = dolfin.Function(self.function_space, name="Ta")
        self._projector = Projector(self.function_space)

    @property
    def dLambda(self):
        self._dLambda.vector()[:] = self.lmbda.vector() - self.lmbda_prev.vector()
        return self._dLambda

    @property
    def Aw(self):
        Tot_A = self._parameters["Tot_A"]
        rs = self._parameters["rs"]
        rw = self._parameters["rw"]
        scale_popu_rw = self._parameters["scale_popu_rw"]
        scale_popu_rs = self._parameters["scale_popu_rs"]
        return (
            Tot_A
            * rs
            * scale_popu_rs
            / (rs * scale_popu_rs + rw * scale_popu_rw * (1.0 - (rs * scale_popu_rs)))
        )

    @property
    def As(self):
        return self.Aw

    @property
    def cw(self):
        phi = self._parameters["phi"]
        kuw = self._parameters["kuw"]
        rw = self._parameters["rw"]

        scale_popu_kuw = self._parameters["scale_popu_kuw"]
        scale_popu_rw = self._parameters["scale_popu_rw"]
        return (
            kuw
            * scale_popu_kuw
            * phi
            * (1.0 - (rw * scale_popu_rw))
            / (rw * scale_popu_rw)
        )

    @property
    def cs(self):
        phi = self._parameters["phi"]
        kws = self._parameters["kws"]
        rs = self._parameters["rs"]
        rw = self._parameters["rw"]
        scale_popu_kws = self._parameters["scale_popu_kws"]
        scale_popu_rw = self._parameters["scale_popu_rw"]
        scale_popu_rs = self._parameters["scale_popu_rs"]
        return (
            kws
            * scale_popu_kws
            * phi
            * rw
            * scale_popu_rw
            * (1.0 - (rs * scale_popu_rs))
            / (rs * scale_popu_rs)
        )

    def update_Zetas(self):
        self._Zetas.vector()[:] = _Zeta(
            self.Zetas_prev.vector(),
            self.As,
            self.cs,
            self.dLambda.vector(),
            self.dt,
            self._scheme,
        )

    @property
    def Zetas(self):
        return self._Zetas

    def update_Zetaw(self):
        self._Zetaw.vector()[:] = _Zeta(
            self.Zetaw_prev.vector(),
            self.Aw,
            self.cw,
            self.dLambda.vector(),
            self.dt,
            self._scheme,
        )

    @property
    def Zetaw(self):
        return self._Zetaw

    @property
    def dt(self) -> float:
        from .setup_models import TimeStepper

        return TimeStepper.ns2ms(self.t - self._t_prev)

    @property
    def t(self) -> float:
        return self._t

    def start_time(self, t):
        self._t_prev = t
        self._t = t

    def update_time(self, t):
        self._t_prev = self.t
        self._t = t

    def update_prev(self):
        self.Zetas_prev.vector()[:] = self.Zetas.vector()
        self.Zetaw_prev.vector()[:] = self.Zetaw.vector()
        self.lmbda_prev.vector()[:] = self.lmbda.vector()
        self._projector.project(self.Ta_current, self.Ta)

    @property
    def Ta(self):
        Tref = self._parameters["Tref"]
        rs = self._parameters["rs"]
        scale_popu_Tref = self._parameters["scale_popu_Tref"]
        scale_popu_rs = self._parameters["scale_popu_rs"]
        Beta0 = self._parameters["Beta0"]

        _min = ufl.min_value
        _max = ufl.max_value
        if isinstance(self.lmbda, (int, float)):
            _min = min
            _max = max
        lmbda = _min(1.2, self.lmbda)
        h_lambda_prima = 1.0 + Beta0 * (lmbda + _min(lmbda, 0.87) - 1.87)
        h_lambda = _max(0, h_lambda_prima)

        return (
            h_lambda
            * (Tref * scale_popu_Tref / (rs * scale_popu_rs))
            * (self.XS * (self.Zetas + 1.0) + self.XW * self.Zetaw)
        )

    def Wactive(self, F, **kwargs):
        """Active stress energy"""

        C = F.T * F
        f = F * self.f0
        self._projector.project(self.lmbda, dolfin.sqrt(f**2))
        self.update_Zetas()
        self.update_Zetaw()
        return pulse.material.active_model.Wactive_transversally(
            Ta=self.Ta,
            C=C,
            f0=self.f0,
            eta=self.eta,
        )
